"""Module that supports barycenter-based flash matching."""

import numpy as np

from spine.math.distance import cdist


class BarycenterFlashMatcher:
    """Matches interactions and flashes by matching the charge barycenter of
    TPC interactions with the light barycenter of optical flashes.
    """

    # List of valid matching methods
    _match_methods = ("threshold", "best")

    def __init__(
        self,
        match_method="threshold",
        dimensions=(1, 2),
        charge_weighted=False,
        time_window=None,
        first_flash_only=False,
        min_inter_size=None,
        min_flash_pe=None,
        match_distance=None,
    ):
        """Initalize the barycenter flash matcher.

        Parameters
        ----------
        match_method: str, default 'distance'
            Matching method (one of 'threshold' or 'best')
            - 'threshold': If the two barycenters are within some distance, match
            - 'best': For each flash, pick the best matched interaction
        dimensions: Tuple[int], default (1, 2)
            Dimensions involved in the distance computation
        charge_weighted : bool, default False
            Use interaction pixel charge information to weight the centroid
        time_window : List, optional
            List of [min, max] values of optical flash times to consider
        first_flash_only : bool, default False
            Only try to match the first flash in the time window
        min_inter_size : int, optional
            Minimum number of voxel in an interaction to consider it
        min_flash_pe : float, optional
            Minimum number of total PE in a flash to consider it
        match_distance : float, optional
            If a threshold is used, specifies the acceptable distance
        """
        # Check validity of key parameters
        if match_method not in self._match_methods:
            raise ValueError(
                "Barycenter flash matching method not recognized: "
                f"{match_method}. Must be one of {self._match_methods}."
            )

        if match_method == "threshold":
            assert match_distance is not None, (
                "When using the `threshold` method, must specify the "
                "`match_distance` parameter."
            )

        # Store the flash matching parameters
        self.match_method = match_method
        self.dims = dimensions
        self.charge_weighted = charge_weighted
        self.time_window = time_window
        self.first_flash_only = first_flash_only
        self.min_inter_size = min_inter_size
        self.min_flash_pe = min_flash_pe
        self.match_distance = match_distance

    def get_matches(self, interactions, flashes):
        """Makes [interaction, flash] pairs that have compatible barycenters.

        Parameters
        ----------
        interactions : List[Union[RecoInteraction, TruthInteraction]]
            List of interactions
        flashes : List[Flash]
            List of optical flashes

        Returns
        -------
        List[[Interaction, larcv.Flash, float]]
            List of [interaction, flash, distance] triplets
        """
        # Restrict the flashes to those that fit the selection criteria.
        # Skip if there are no valid flashes
        if self.time_window is not None:
            t1, t2 = self.time_window
            flashes = [f for f in flashes if (f.time > t1 and f.time < t2)]

        if self.min_flash_pe is not None:
            flashes = [f for f in flashes if f.total_pe > self.min_flash_pe]

        if len(flashes) == 0:
            return []

        # If requested, restrict the list of flashes to match to the first one
        if self.first_flash_only:
            flashes = [flashes[0]]

        # Restrict the interactions to those that fit the selection criterion.
        # Skip if there are no valid interactions
        if self.min_inter_size is not None:
            interactions = [
                inter for inter in interactions if inter.size > self.min_inter_size
            ]

        if len(interactions) == 0:
            return []

        # Get the flash centroids
        op_centroids = np.empty((len(flashes), len(self.dims)))
        op_widths = np.empty((len(flashes), len(self.dims)))
        for i, f in enumerate(flashes):
            op_centroids[i] = f.center[self.dims]
            op_widths[i] = f.width[self.dims]

        # Get interactions centroids
        int_centroids = np.empty((len(interactions), len(self.dims)))
        for i, inter in enumerate(interactions):
            if not self.charge_weighted:
                int_centroids[i] = np.mean(inter.points[:, self.dims], axis=0)

            else:
                int_centroids[i] = np.sum(
                    inter.depositions * inter.points[:, self.dims], axis=0
                )
                int_centroids[i] /= np.sum(inter.depositions)

        # Compute the flash to interaction distance matrix
        dist_mat = cdist(op_centroids, int_centroids)

        # Produce matches
        matches = []
        if self.match_method == "best":
            # For each flash, select the best match, save attributes
            for i, f in enumerate(flashes):
                best_match = np.argmin(dist_mat[i])
                dist = dist_mat[i, best_match]
                if self.match_distance is not None and dist > self.match_distance:
                    continue
                matches.append((interactions[best_match], f, dist))

        elif self.match_method == "threshold":
            # Find all compatible pairs
            valid_pairs = np.vstack(np.where(dist_mat <= self.match_distance)).T
            matches = [
                (interactions[j], flashes[i], dist_mat[i, j]) for i, j in valid_pairs
            ]

        return matches



"""Detector-geometry-based reconstruction module."""

from spine.geo import GeoManager
from spine.post.base import PostBase
from spine.utils.globals import PID_LABELS

__all__ = ["ContainmentProcessor", "FiducialProcessor"]


class ContainmentProcessor(PostBase):
    """Check whether a fragment, particle or interaction is contained.

    This module checks whether the object comes within some distance of the
    boundaries of the detector and assign the `is_contained` attribute
    accordingly.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "containment"

    # Alternative allowed names of the post-processor
    aliases = ("time_containment", "check_containment", "check_time_containment")

    def __init__(
        self,
        margin,
        cathode_margin=None,
        mode="module",
        allow_multi_module=False,
        logical=False,
        exclude_pids=None,
        min_particle_sizes=0,
        obj_type=("particle", "interaction"),
        truth_point_mode="points",
        run_mode="both",
    ):
        """Initialize the containment conditions.

        If the `source` method is used, the cut will be based on the source of
        the point cloud, i.e. if a point cloud was produced by TPCs i and j, it
        must be contained within the volume bound by the set of TPCs i and j,
        and whichever volume is present between them.

        Parameters
        ----------
        margin : Union[float, List[float], np.array]
            Minimum distance from a detector wall to be considered contained:
            - If float: distance buffer is shared between all 6 walls
            - If [x,y,z]: distance is shared between pairs of falls facing
              each other and perpendicular to a shared axis
            - If [[x_low,x_up], [y_low,y_up], [z_low,z_up]]: distance is
              specified individually of each wall.
        cathode_margin : float, optional
            If specified, sets a different margin for the cathode boundaries
        mode : str, default 'module'
            Containement criterion (one of 'global', 'module', 'tpc'):
            - If 'tpc', makes sure it is contained within a single tpc
            - If 'module', makes sure it is contained within a single module
            - If 'detector', makes sure it is contained within the
              outermost walls
            - If 'source', use the origin of voxels to determine which TPC(s)
              contributed to them, and define volumes accordingly
            - If 'meta', use the metadata range as a basis for containment.
              Note that this does not guarantee containment within the detector.
        allow_multi_module : bool, default False
            If `True`, allow particles/interactions to span multiple modules
        logical : bool, default False
            If `True`, the module is checking for logical containment, i.e. that
            particles/interactions do not span outside of the drift window, with
            padding defined by the margin/cathode margin parameters
        exclude_pids : List[int], optional
            When checking interaction containment, ignore particles which
            have a species specified by this list
        min_particle_sizes : Union[int, dict], default 0
            When checking interaction containment, ignore particles below the
            size (in voxel count) specified by this parameter. If specified
            as a dictionary, it maps a specific particle type to its own cut.
        """
        # Initialize the parent class
        super().__init__(obj_type, run_mode, truth_point_mode)

        # Initialize the geometry, if needed
        if mode != "meta":
            self.use_meta = False
            self.geo = GeoManager.get_instance()
            self.cont_def = self.geo.define_containment_volumes(
                margin, cathode_margin, mode
            )

        else:
            self.update_keys({"meta": True})
            self.use_meta = True
            self.margin = margin

        # If G4 truth points are used for truth containment, use dedicated check
        self.truth_cont_def = None
        if "g4" in self.truth_point_mode and mode != "meta":
            truth_mode = "module" if mode != "detector" else mode
            self.geo = GeoManager.get_instance()
            self.truth_cont_def = self.geo.define_containment_volumes(
                margin, mode=truth_mode
            )

        # Store parameters
        self.allow_multi_module = allow_multi_module
        self.exclude_pids = exclude_pids

        # Check which type of containment check is being operated
        assert (
            not logical or mode == "source"
        ), "For logical containment, must use the individual point sources."
        self.attr = "is_contained" if not logical else "is_time_contained"

        # Store the particle size thresholds in a dictionary
        self.min_particle_sizes = {}
        for pid in PID_LABELS:
            if isinstance(min_particle_sizes, int):
                self.min_particle_sizes[pid] = min_particle_sizes
            elif pid in min_particle_sizes:
                self.min_particle_sizes[pid] = min_particle_sizes[pid]
            elif "default" in min_particle_sizes:
                self.min_particle_sizes[pid] = min_particle_sizes["default"]
            else:
                self.min_particle_sizes[pid] = 0

    def process(self, data):
        """Check the containment of all objects in one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over particle objects
        for k in self.fragment_keys + self.particle_keys:
            for obj in data[k]:
                # Make sure the particle coordinates are expressed in cm
                self.check_units(obj)

                # Get point coordinates
                points = self.get_points(obj)
                if len(points) == 0:
                    setattr(obj, self.attr, True)
                    continue

                # Check particle containment against detector/meta
                if self.use_meta:
                    obj.is_contained = (
                        points > (data["meta"].lower + self.margin)
                    ).all() and (points < (data["meta"].upper - self.margin)).all()

                else:
                    if not obj.is_truth or self.truth_cont_def is None:
                        sources = self.get_sources(obj)
                        setattr(
                            obj,
                            self.attr,
                            self.geo.check_containment(
                                self.cont_def,
                                points,
                                sources,
                                allow_multi_module=self.allow_multi_module,
                            ),
                        )
                    else:
                        setattr(
                            obj,
                            self.attr,
                            self.geo.check_containment(
                                self.truth_cont_def,
                                points,
                                allow_multi_module=self.allow_multi_module,
                            ),
                        )

        # Loop over interaction objects
        for k in self.interaction_keys:
            for inter in data[k]:
                # Check that all the particles in the interaction are contained
                setattr(inter, self.attr, True)
                for part in inter.particles:
                    if not getattr(part, self.attr):
                        # Do not check for particles with excluded PID
                        if (
                            self.exclude_pids is not None
                            and part.pid in self.exclude_pids
                        ):
                            continue

                        # Do not account for particles below a certain size
                        if (
                            part.pid > -1
                            and part.size < self.min_particle_sizes[part.pid]
                        ):
                            continue

                        setattr(inter, self.attr, False)
                        break


class FiducialProcessor(PostBase):
    """Check that an interaction vertex is in a user-defined fiducial volume.

    The fiducial volume is defined as a margin distances from each of the
    detector walls.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "fiducial"

    # Alternative allowed names of the post-processor
    aliases = ("check_fiducial",)

    # Set of post-processors which must be run before this one is
    _upstream = ("vertex",)

    # List of valid ways to fetch a true vertex
    _truth_vertex_modes = ("vertex", "reco_vertex")

    def __init__(
        self,
        margin,
        cathode_margin=None,
        mode="module",
        run_mode="both",
        truth_vertex_mode="vertex",
    ):
        """Initialize the fiducial conditions.

        Parameters
        ----------
        margin : Union[float, List[float], np.array]
            Minimum distance from a detector wall to be considered contained:
            - If float: distance buffer is shared between all 6 walls
            - If [x,y,z]: distance is shared between pairs of falls facing
              each other and perpendicular to a shared axis
            - If [[x_low,x_up], [y_low,y_up], [z_low,z_up]]: distance is
              specified individually of each wall.
        cathode_margin : float, optional
            If specified, sets a different margin for the cathode boundaries
        mode : str, default 'module'
            Containement criterion (one of 'global', 'module', 'tpc'):
            - If 'tpc', makes sure it is contained within a single tpc
            - If 'module', makes sure it is contained within a single module
            - If 'detector', makes sure it is contained within the
              outermost walls
            - If 'meta', use the metadata range as a basis for containment.
              Note that this does not guarantee containment within the detector.
        truth_vertex_mode : str, default 'truth_vertex'
             Vertex attribute to use to check containment of true interactions
        """
        # Initialize the parent class
        super().__init__("interaction", run_mode)

        # Initialize the geometry
        if mode != "meta":
            self.use_meta = False
            self.geo = GeoManager.get_instance()
            self.fid_def = self.geo.define_containment_volumes(
                margin, cathode_margin, mode, include_limits=False
            )

        else:
            self.update_keys({"meta": True})
            self.use_meta = True
            self.margin = margin

        # Store the true vertex source
        assert truth_vertex_mode in self._truth_vertex_modes, (
            f"`truth_vertex_mode not recognized: `{truth_vertex_mode}`. "
            f"Must be one one of {self._truth_vertex_modes}."
        )
        self.truth_vertex_mode = truth_vertex_mode

    def process(self, data):
        """Check the fiducial status of all interactions in one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over interaction objects
        for k in self.interaction_keys:
            for inter in data[k]:
                # Make sure the interaction coordinates are expressed in cm
                self.check_units(inter)

                # Get point coordinates
                vertex = self.get_vertex(inter).reshape(-1, 3)

                # Check containment
                if not self.use_meta:
                    inter.is_fiducial = self.geo.check_containment(self.fid_def, vertex)
                else:
                    inter.is_fiducial = (
                        vertex > (data["meta"].lower + self.margin)
                    ).all() and (vertex < (data["meta"].upper - self.margin)).all()

    def get_vertex(self, inter):
        """Get a certain pre-defined vertex attribute of an interaction.

        The :class:`TruthInteraction` vertexes are obtained using the
        `truth_vertex_mode` attribute of the class.

        Parameters
        ----------
        obj : InteractionBase
            Interaction object

        Returns
        -------
        np.ndarray
            (N) Interaction vertex
        """
        if not inter.is_truth:
            return inter.vertex
        else:
            return getattr(inter, self.truth_vertex_mode)



"""Used to define which dataset entries to load at each iteration"""

import time

import numpy as np
from torch.utils.data import Sampler
from torch.utils.data.distributed import DistributedSampler

__all__ = [
    "SequentialBatchSampler",
    "RandomSequenceBatchSampler",
    "BootstrapBatchSampler",
]


class AbstractBatchSampler(Sampler):
    """Abstract sampler class.

    Samplers that inherit from this class should work out of the box.
    Just define the __len__ and __iter__ functions. __init__ defines
    self.num_samples and self.batch_size as well as a self._random
    RNG, if needed.
    """

    def __init__(self, dataset, batch_size, seed=None, drop_last=True):
        """Check and store the values passed to the initializer,
        set the seeds appropriately.

        Parameters
        ----------
        dataset : torch.utils.data.Dataset
            Dataset to sampler from
        batch_size : int
            Number of samples to load per iteration
        seed : int, optional
            Seed to use for random sampling
        drop_last: bool, default True
            If `True`, drop the last batch to make the number of entries a
            multiple of the batch_size (if needed)
        """
        # Initialize parent class
        # TODO: `data_source` id deprecated, will need to remove with newer
        # version of pytorch (2.2.0 and above)
        super().__init__(data_source=None)

        # Initialize the random number generator with a seed
        if seed is None:
            seed = int(time.time())
        else:
            assert isinstance(
                seed, int
            ), f"The sampler seed must be an integer, got: {seed}."

        self._random = np.random.RandomState(seed=seed)  # pylint: disable=E1101

        # Check that the batch_size is a sensible value
        self.batch_size = batch_size
        if batch_size < 1:
            raise ValueError("The `batch_size` must be a positive non-zero integer.")

        # Process the number of samples
        self.num_samples = len(dataset)
        self.drop_last = drop_last
        if self.num_samples < 1:
            raise ValueError("The dataset must have at least on entry to sample from.")
        if drop_last:
            if self.num_samples < self.batch_size:
                raise ValueError(
                    "The dataset does not have enough samples "
                    f"({self.num_samples}) to produce a complete batch "
                    f"({batch_size})."
                )
            self.num_samples -= self.num_samples % self.batch_size

    def __len__(self):
        """Provides the full length of the sampler.

        The length of the sampler can differ from the number of elements in
        the underlying dataset, if the last batch is smaller than the requested
        size and is dropped.

        Returns
        -------
        int
            Total number of entries to sample
        """
        return self.num_samples

    def __iter__(self):
        """Placeholder to be overridden by children classes."""
        raise NotImplementedError


class SequentialBatchSampler(AbstractBatchSampler):
    """Samples batches sequentially within the dataset."""

    name = "sequential"

    def __iter__(self):
        """Iterates over sequential batches of data."""
        order = np.arange(self.num_samples, dtype=int)
        return iter(order)


class RandomSequenceBatchSampler(AbstractBatchSampler):
    """Samples sequential batches randomly within the dataset."""

    name = "random_sequence"

    def __iter__(self):
        """Iterates over sequential batches of data randomly located
        in the dataset.
        """
        # Pick a general offset and produce sequence starts with respect to it.
        # Introducing this random offset ensures that the data is not
        # systematically batched in the same way
        offset = self._random.randint(0, self.batch_size)
        starts = np.arange(
            offset, offset + self.num_samples, self.batch_size, dtype=int
        )

        # Randomly pick the starts
        self._random.shuffle(starts)

        # Produce a sequence for each start
        batches = [
            np.arange(start, start + self.batch_size, dtype=int) for start in starts
        ]
        indices = np.concatenate(batches)

        # Wrap indexes around if the offset is non-zero
        indices = indices % self.num_samples

        return iter(indices)


class BootstrapBatchSampler(AbstractBatchSampler):
    """Sampler used for bootstrap sampling of the entire dataset.

    This is particularly useful for training an ensemble of networks
    (bagging).
    """

    name = "bootstrap"

    def __iter__(self):
        """Iterates over bootstrapped batches of data randomly picked
        from the dataset.
        """
        max_id = self.num_samples + 1 - self.batch_size
        starts = np.arange(0, max_id, self.batch_size, dtype=int)
        bootstrap_indices = np.random.choice(
            np.arange(self.num_samples), self.num_samples
        )
        batches = [
            bootstrap_indices[np.arange(start, start + self.batch_size, dtype=int)]
            for start in starts
        ]

        return iter(np.concatenate(batches))


class DistributedProxySampler(DistributedSampler):
    """Sampler that restricts data loading to a subset of input sampler indices.

    It is especially useful in conjunction with
    :class:`torch.nn.parallel.DistributedDataParallel`. In such case, each
    process can pass a DistributedSampler instance as a DataLoader sampler,
    and load a subset of the original dataset that is exclusive to it.
    """

    def __init__(self, sampler, num_replicas, rank):
        """Convert a basic sampler to an instance of a distributed sampler.

        Parameters
        ----------
        sampler : Sampler
            Input torch sampler
        num_replicas : int
            Number of distributed samplers running concurrently
        rank : int
            Rank of the current sampler

        Notes
        -----
        Input sampler is assumed to be of constant size.
        """
        # Make sure the batch_size is a multiple of the number of replicas
        assert sampler.batch_size % num_replicas == 0, (
            f"The `batch_size` ({sampler.batch_size}) must be a multiple "
            f"of the number of replicas ({num_replicas}) in the "
            "distributed training process."
        )

        # Initialiaze the parent distributed sampler
        super().__init__(
            sampler,
            num_replicas=num_replicas,
            rank=rank,
            drop_last=sampler.drop_last,
            shuffle=False,
        )

        # Store the underlying sampler and its parameters
        self.sampler = sampler
        self.batch_size = sampler.batch_size

    def __iter__(self):
        """Overrides the basic iterator with one that takes into account
        the number of replicas and the rank of the sampler.
        """
        # Fetch the list of non-distributed indices
        indices = list(self.sampler)

        # If the number of entries is not a multiple of the number of replicas,
        # must pad the end.
        if not self.drop_last:
            padding_size = self.total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[
                    :padding_size
                ]

        assert len(indices) == self.total_size

        # Subsample by keeping the indices sequential between each minibatch.
        # This is crucial to preserve contiguousness in sequential samplers.
        minibatch_size = self.batch_size / self.num_replicas
        ranks = (np.arange(self.total_size) // minibatch_size) % self.num_replicas

        indices = np.array(indices, dtype=int)[ranks == self.rank]
        assert len(indices) == self.num_samples

        return iter(indices)



import MinkowskiEngine as ME
import torch
import torch.nn as nn


class ImageClassificationWrapper(nn.Module):

    def __init__(self, model):
        super(ImageClassificationWrapper, self).__init__()
        self.model = model

    def forward(self, x, output_name="logits"):

        res = self.model.forward(x)
        return res[output_name][0]



"""Class to build all require representations."""

from collections import OrderedDict, defaultdict

import numpy as np

from spine.utils.globals import COORD_COLS, VALUE_COL

from .fragment import FragmentBuilder
from .interaction import InteractionBuilder
from .particle import ParticleBuilder


class BuildManager:
    """Manager which constructs data representations based on the chain output.

    Takes care of two scenarios:
      - Interpret the raw output of the reconstruction chain
      - Load up existing objects stored as dictionaries
    """

    # List of recognized run modes
    _run_modes = ("reco", "truth", "both", "all")

    # List of recognized units
    _units = ("cm", "px")

    # Name of input data products needed to build representations. These names
    # are not set in stone; they can be set in the configuration
    _sources = (
        ("data_tensor", ("data_adapt", "data")),
        ("label_tensor", ("clust_label",)),
        ("label_adapt_tensor", ("clust_label_adapt",)),
        ("label_g4_tensor", ("clust_label_g4",)),
        ("depositions_q_label", ("charge_label",)),
        ("graph_label", ("graph_label",)),
        ("sources", ("sources_adapt", "sources")),
        ("sources_label", ("sources_label",)),
        ("particles", ("particles",)),
        ("neutrinos", ("neutrinos",)),
        ("flashes", ("flashes",)),
        ("crthits", ("crthits",)),
    )

    def __init__(
        self,
        fragments,
        particles,
        interactions,
        mode="both",
        units="cm",
        sources=None,
        lite=False,
    ):
        """Initializes the build manager.

        Parameters
        ----------
        fragments : bool
            Build/load RecoFragment/TruthFragment objects
        particles : bool
            Build/load RecoParticle/TruthParticle objects
        interactions : bool
            Build/load RecoInteraction/TruthInteraction objects
        mode : str, default 'both'
            Whether to construct reconstructed objects, true objects or both
        sources : Dict[str, str], optional
            Dictionary which maps the necessary data products onto a name
            in the input/output dictionary of the reconstruction chain.
        lite : bool, default False
            If `True`, the objects being loaded are lite and do not map
            to long-form attributes. Simply load the matches.
        """
        # Check on the mode, store it
        assert (
            mode in self._run_modes
        ), f"Run mode not recognized: {mode}. Must be one {self._run_modes}"
        self.mode = mode

        # Check on the units, store them
        assert (
            units in self._units
        ), f"Units not recognized: {units}. Must be one {self._units}"
        self.units = units

        # If custom sources are provided, update the tuple
        if sources is not None:
            sources_dict = dict(self._sources)
            for key, value in sources.items():
                assert key in sources_dict, (
                    "Unexpected data product specified in `sources`: "
                    f"{key}. Should be one of {list(sources_dict.keys())}."
                )
                if isinstance(value, str):
                    sources_dict[key] = (value,)
                else:
                    sources_dict[key] = tuple(value)

            self._sources = tuple(sources_dict.items())

        # Initialize the builders
        self.builders = OrderedDict()
        if fragments:
            self.builders["fragment"] = FragmentBuilder(mode, units)
        if particles:
            self.builders["particle"] = ParticleBuilder(mode, units)
        if interactions:
            assert particles, (
                "Interactions are built from particles. If `interactions` "
                "is True, so must `particles` be."
            )
            self.builders["interaction"] = InteractionBuilder(mode, units)

        # Store whether to load the long-form attributes or not
        self.lite = lite

    def __call__(self, data):
        """Build the representations for one entry.

        Parameters
        ----------
        data : dict
            Dictionary of input data and model outputs

        Notes
        -----
        Modifies the data dictionary in place.
        """
        # If this is the first time the builders are called, build
        # the objects shared between fragments/particles/interactions
        load = True
        if not self.lite and "points" not in data and "points_label" not in data:
            load = False
            if np.isscalar(data["index"]):
                sources = self.build_sources(data)
            else:
                sources = defaultdict(list)
                for entry in range(len(data["index"])):
                    sources_e = self.build_sources(data, entry)
                    for key, val in sources_e.items():
                        sources[key].append(val)

            data.update(**sources)

        # Loop over builders
        for name, builder in self.builders.items():
            # Build representations
            builder(data)

            # Generate match pairs from stored matches
            if load and self.mode in ["both", "all"]:
                if np.isscalar(data["index"]):
                    match_dict = self.load_match_pairs(data, name)
                else:
                    match_dict = defaultdict(list)
                    for entry in range(len(data["index"])):
                        match_dict_e = self.load_match_pairs(data, name, entry)
                        for key, val in match_dict_e.items():
                            match_dict[key].append(val)

                data.update(**match_dict)

    def build_sources(self, data, entry=None):
        """Construct the reference coordinate and value tensors used by
        all the representations built by the module.

        These objects should be stored along with the constructed objects
        if the objects are to be loaded later on.

        Parameters
        ----------
        data : dict
            Dictionary of input data and model outputs
        entry : int, optional
            Entry number
        """
        # Fetch the orginal sources
        sources = {}
        for key, alt_keys in self._sources:
            for alt in alt_keys:
                if alt in data:
                    sources[key] = data[alt]
                    if entry is not None:
                        sources[key] = data[alt][entry]
                    break

        # Build aditional information
        update = {}
        if self.mode != "truth" or "label_adapt_tensor" in sources:
            update["points"] = sources["data_tensor"][:, COORD_COLS]
            update["depositions"] = sources["data_tensor"][:, VALUE_COL]

            if "sources" in sources:
                update["sources"] = sources["sources"].astype(int)

        if self.mode != "reco":
            update["label_tensor"] = sources["label_tensor"]
            update["points_label"] = sources["label_tensor"][:, COORD_COLS]
            update["depositions_label"] = sources["label_tensor"][:, VALUE_COL]

            if "depositions_q_label" in sources:
                update["depositions_q_label"] = sources["depositions_q_label"][
                    :, VALUE_COL
                ]

            if "label_adapt_tensor" in sources:
                update["label_adapt_tensor"] = sources["label_adapt_tensor"]
                update["depositions_label_adapt"] = sources["label_adapt_tensor"][
                    :, VALUE_COL
                ]

            if "label_g4_tensor" in sources:
                update["label_g4_tensor"] = sources["label_g4_tensor"]
                update["points_g4"] = sources["label_g4_tensor"][:, COORD_COLS]
                update["depositions_g4"] = sources["label_g4_tensor"][:, VALUE_COL]

            if "sources_label" in sources:
                update["sources_label"] = sources["sources_label"].astype(int)

        # If provided, etch the point attributes to check their units
        for obj in ["fragment", "particle"]:
            for key in [f"{obj}_start_points", f"{obj}_end_points"]:
                if key in data:
                    update[key] = data[key]
                    if entry is not None:
                        update[key] = update[key][entry]

        # Convert everything to the proper units once and for all
        if self.units != "px":
            # Fetch metadata
            assert "meta" in data, "Must provide metadata to build objects in cm."

            meta = data["meta"][entry] if entry is not None else data["meta"]
            for key in update:
                if "points" in key and key in update:
                    if key in update:
                        update[key] = meta.to_cm(np.copy(update[key]), center=True)

            for key in ["particles", "neutrinos"]:
                if key in sources:
                    update[key] = sources[key]
                    for obj in sources[key]:
                        if obj.units != self.units:
                            obj.to_cm(meta)

        return update

    @staticmethod
    def load_match_pairs(data, name, entry=None):
        """Generate lists of matched object pairs from stored matches.

        Parameters
        ----------
        data : dict
            Dictionary of input data and model outputs
        name : str
            Object type name
        entry : int, optional
            Entry number
        """
        # Initialize the name of the match lists
        prefix = f"{name}_matches"

        # Create match pairs in both directions (true to reco and vice versa)
        result = {}
        for source, target in [("reco", "truth"), ("truth", "reco")]:
            # Fetch the lists of objects to match
            sources = data[f"{source}_{name}s"]
            targets = data[f"{target}_{name}s"]
            if entry is not None:
                sources, targets = sources[entry], targets[entry]

            # Loop
            suffix = f"{source[0]}2{target[0]}"
            match_key = f"{prefix}_{suffix}"
            match_overlap_key = f"{match_key}_overlap"
            result[match_key] = []
            result[match_overlap_key] = []
            for obj in sources:
                if not obj.is_matched:
                    # If no match is found, give an empty value to the match
                    result[match_key].append((obj, None))
                    result[match_overlap_key].append(-1.0)

                else:
                    # If a match is found, the first is always the best match
                    best_match = obj.match_ids[0]
                    result[match_key].append((obj, targets[best_match]))
                    result[match_overlap_key].append(obj.match_overlaps[0])

        return result



"""Module with methods to overlay multiple events."""

from dataclasses import dataclass
from warnings import warn

import numpy as np

from ..core.parse.clean_data import clean_sparse_data
from ..core.parse.data import ParserObjectList, ParserTensor

__all__ = ["Overlayer"]


class Overlayer:
    """Generic class to produce data overlays.

    This class supports 3 image overlay modes:
    - 'constant' will produce overlays with a constant multiplicity;
    - 'uniform' will produce overlays with multiplicities, M_i, sampled from a
      uniform distribution such that, for a batch size B, \Sum_i M_i = B;
    - 'poisson' will produce overlays with multiplicities, M_i, sampled from a
      Poisson distribution of mean set by 'multiplicity'. For a batch size B,
      the multiplicities are set such that \Sum_i M_i = B.
    """

    # List of recognized overlay modes
    _modes = ("constant", "uniform", "poisson")

    def __init__(self, data_types, methods, multiplicity, mode="constant"):
        """Store the overlay parameters.

        Parameters
        ----------
        data_types : Dict[str, str]
            Types of data returned by the upstream parsers
        methods : Dict[str, str]
            Maps data products onto overlay methods
        multiplicity : int
            Number of images to stack in the overlay
        mode : str, default 'constant'
            Overlay mode (one of 'constant', 'uniform' or 'poisson')
        """
        # Check that the overlay mode is recognized
        assert mode in self._modes, (
            f"Overlay mode not recognized: {mode}. Must be one of " f"{self._modes}."
        )
        self.mode = mode

        # Check that multiplicity is sensible
        assert (
            multiplicity > 0
        ), "Overlay multiplicity should be a non-zero positive integer."
        self.multiplicity = multiplicity

        # Store the data types and methods
        self.data_types = data_types
        self.methods = methods

    def __call__(self, batch):
        """Given a batch of data, provides an overlay batching and modifies
        the data in place to avoid indexing conflicts.

        Parameters
        ----------
        batch : List[Dict]
            List of dictionaries of parsed information, one per event. Each
            dictionary matches one data key to one event-worth of parsed data.

        Returns
        -------
        List[Dict]
            Overlayed list of dictionaries of parsed information, one per overlay.
        """
        # Fetch the batch size, build an overlap map
        batch_size = len(batch)
        overlay_ids = self.get_assignments(batch_size)

        # Loop over the unique overlay indexes
        overlay_batch = []
        _, splits = np.unique(overlay_ids, return_index=True)
        indexes = np.split(np.arange(batch_size), splits[1:])
        for overlay_id, index in enumerate(indexes):
            # If there is only a single index in the overlay, nothing to do
            if len(index) < 2:
                overlay_batch.append(batch[index[0]])
                continue

            # Loop over the keys to overlay
            overlay = {}
            for key, data_type in self.data_types.items():
                # Dispatch and fill the overlay
                if data_type == "scalar":
                    # Check whether scalars can be harmonized
                    overlay[key] = self.merge_scalars(batch, key, index)

                elif data_type == "object":
                    # Check that objects are compatible when overlaying
                    overlay[key] = self.merge_objects(batch, key, index)

                elif data_type == "object_list":
                    # Offset object list index attributes if needed
                    overlay[key] = self.cat_objects(batch, key, index)

                elif data_type == "tensor":
                    # Stack tensors, offset index columns if needed
                    overlay[key] = self.stack_tensors(batch, key, index)

            # Add overlay to the batch
            overlay_batch.append(overlay)

        return overlay_batch

    def get_assignments(self, batch_size):
        """Given a data product count, produce batch assignments.

        Parameters
        ----------
        batch_size : int
            Number of entries in the batch

        Returns
        -------
        np.ndarray
            Overlay ID assignments
        """
        # Dispatch
        if self.mode == "constant":
            # Uniform multiplicity of overlays
            if batch_size % self.multiplicity != 0:
                warn(
                    f"The overlay multiplicity ({self.multiplicity}) is not a "
                    "divider of the batch size ({batch_size}). The overlay "
                    "multiplicity will not be uniform."
                )

            overlay_ids = np.arange(batch_size, dtype=int) // self.multiplicity

        elif self.mode in ["poisson", "uniform"]:
            # Sample from a Poisson distribution until it adds up to the batch size
            overlay_ids = np.empty(batch_size, dtype=int)
            idx, total = 0, 0
            while total < batch_size:
                # Sample distribution
                if self.mode == "poisson":
                    sample = np.random.poisson(self.multiplicity)
                else:
                    sample = np.random.randint(1, self.multiplicity + 1)

                # Assign overlay indices
                if sample > 0:
                    overlay_ids[total : total + sample] = idx
                    idx += 1
                    total += sample

        # Return
        return overlay_ids

    def merge_scalars(self, batch, key, index):
        """Merge scalars into one per overlay.

        Parameters
        ----------
        batch : List[Dict]
            List of dictionaries of parsed information, one per event. Each
            dictionary matches one data key to one event-worth of parsed data.
        key : str
            Scalar data product key
        index : np.ndarray
            List of indexes to merge into an overlay

        Returns
        -------
        object
            Single scalar for the batch
        """
        scalars = np.array([batch[idx][key] for idx in index])
        if self.methods[key] in ["first", "match"]:
            # Make sure that all scalars match within the overlay, if needed
            if self.methods[key] == "match":
                if not np.all(scalars[1:] == scalars[0]):
                    raise ValueError(
                        f"The scalar values to overlay do not match for {key}."
                    )

            return scalars[0]

        elif self.methods[key] == "sum":
            # Sum the values within each overlay
            return np.sum(scalars)

        elif self.methods[key] == "cat":
            # Concatenate the scalars in a single array (type change)
            return scalars

        else:
            if self.methods[key] is None:
                raise ValueError(f"Scalar overlay method not specified for {key}.")

            raise ValueError(
                f"Scalar overlay method not recognized: {self.methods[key]}. "
                "Must be one of 'first', 'match' or 'sum'."
            )

    def merge_objects(self, batch, key, index):
        """Merge objects into one per overlay.

        Parameters
        ----------
        batch : List[Dict]
            List of dictionaries of parsed information, one per event. Each
            dictionary matches one data key to one event-worth of parsed data.
        key : str
            Object data product key
        index : np.ndarray
            List of indexes to merge into an overlay

        Returns
        -------
        object
            Single object for the batch
        """
        objects = [batch[idx][key] for idx in index]
        if self.methods[key] in ["first", "match"]:
            # Make sure that all objects match within the overlay, if needed
            if self.methods[key] == "match":
                if not np.all([obj == objects[0] for obj in objects]):
                    raise ValueError(f"The objects to overlay do not match for {key}.")

            return objects[0]

        elif self.methods[key] == "cat":
            # Concatenate the objects in a single list (type change)
            return ParserObjectList(objects, default=objects[0])

        else:
            if self.methods[key] is None:
                raise ValueError(f"Object overlay method not specified for {key}.")

            raise ValueError(
                f"Object overlay method not recognized: {self.methods[key]}. "
                "Must be one of 'first' or 'match'."
            )

    def cat_objects(self, batch, key, index):
        """Concatenate object lists into one, offset index attributes if needed.

        Parameters
        ----------
        batch : List[Dict]
            List of dictionaries of parsed information, one per event. Each
            dictionary matches one data key to one event-worth of parsed data.
        key : str
            Object list data product key
        index : np.ndarray
            List of indexes to merge into an overlay

        Returns
        -------
        ObjList
            Concatenated obejct list
        """
        # If the objects in the lists contain indexes, must offset them
        ref_list = batch[index[0]][key]
        shifts = None
        if len(ref_list.default.index_attrs) > 0:
            shifts = ref_list.index_shifts
            for idx in index[1:]:
                # Shift indexes in the objects
                obj_list = batch[idx][key]
                for obj in obj_list:
                    obj.shift_indexes(shifts)

                # Increment shifts
                if not isinstance(shifts, dict):
                    shifts += obj_list.index_shifts
                else:
                    for attr in shifts:
                        shifts[attr] += obj_list.index_shifts[attr]

        # Concatenate and return
        obj_list = []
        for idx in index:
            obj_list.extend(batch[idx][key])

        return ParserObjectList(obj_list, ref_list.default, shifts)

    def stack_tensors(self, batch, key, index):
        """Stack tensors together across an overlay.

        Parameters
        ----------
        batch : List[Dict]
            List of dictionaries of parsed information, one per event. Each
            dictionary matches one data key to one event-worth of parsed data.
        key : str
            Tensor data product key
        index : np.ndarray
            List of indexes to merge into an overlay

        Returns
        -------
        ParserTensor
            Overlayed tensor
        """
        # Define a reference tensor
        ref_data = batch[index[0]][key]

        # Stack coordinates, if present
        coords = None
        if ref_data.coords is not None:
            # Check that the meta data matches between all images (it must)
            if not np.all([batch[idx][key].meta == ref_data.meta for idx in index]):
                raise ValueError("The metadata must match across all overlayed tensor.")
            coords = np.vstack([batch[idx][key].coords for idx in index])

        # If required, offset indexes in the feature tensor
        global_shift, index_shifts = None, None
        if ref_data.global_shift is not None:
            # Shift the whole feature tensor (index tensor)
            global_shift = ref_data.global_shift
            for idx in index[1:]:
                mask = batch[idx][key].features > -1
                batch[idx][key].features[mask] += global_shift
                global_shift += batch[idx][key].global_shift

        elif ref_data.feat_index_cols is not None:
            # Apply offsets to the relevant columns only (mixed features)
            index_shifts = ref_data.index_shifts.copy()
            for idx in index[1:]:
                for i, col in enumerate(ref_data.feat_index_cols):
                    mask = batch[idx][key].features[:, col] > -1
                    batch[idx][key].features[mask, col] += index_shifts[i]
                index_shifts += batch[idx][key].index_shifts

        # Stack features
        if ref_data.global_shift is None:
            features = np.vstack([batch[idx][key].features for idx in index])
        else:
            features = np.concatenate(
                [batch[idx][key].features for idx in index], axis=-1
            )

        # If requested, remove rows corresponding to duplicate coordinates
        if ref_data.remove_duplicates:
            # Check that we have coordinates to make the check
            assert (
                ref_data.coords is not None
            ), "Must provide coordinates to filter duplicates."

            # Filter out duplicates, aggregate features
            coords, features = clean_sparse_data(
                coords,
                features,
                sum_cols=ref_data.feat_sum_cols,
                prec_col=ref_data.feat_prec_col,
                precedence=ref_data.precedence,
            )

        # Returns
        return ParserTensor(
            coords=coords,
            features=features,
            meta=ref_data.meta,
            global_shift=global_shift,
            index_shifts=index_shifts,
            index_cols=ref_data.index_cols,
            sum_cols=ref_data.sum_cols,
            prec_col=ref_data.prec_col,
            precedence=ref_data.precedence,
            feats_only=ref_data.feats_only,
        )



"""Constants used in calibration modules."""

from typing import Dict, List, Optional, Union

import numpy as np

from .database import CalibrationDatabase


class CalibrationConstant:
    """Handles calibration constants which can be either scalar/per-TPC
    or provided via a database.
    """

    def __init__(
        self,
        num_tpcs: int,
        value: Optional[Union[float, List[float]]] = None,
        database: Optional[Union[str, Dict[int, Union[float, List[float]]]]] = None,
    ):
        """Initialize a calibration constant.

        Parameters
        ----------
        num_tpcs : int
            Number of TPCs in the detector
        value : Union[float, List[float]], optional
            Calibration constant (unique or per TPC)
        database : Union[str, Dict[int, Union[float, List[float]]]], optional
            Path to a SQLite db file or dictionary which maps run IDs to calibration
            constants or dictionary which maps run IDs to calibration constants
        """
        # Must provide either a value or a database
        if (value is None) == (database is None):
            raise ValueError("Must provide either a value or a database.")

        # Initialize the gain values
        self.value, self.database = None, None
        if value is not None:
            # If values are provided, make sure they are correct
            if isinstance(value, (list, tuple, np.ndarray)) and len(value) != num_tpcs:
                raise ValueError(
                    "Calibration constant must be a scalar or given per TPC "
                    f"({num_tpcs}). Got {value} instead."
                )
            self.value = self.load_value(value, num_tpcs)

        else:
            # If a database path is provided, load it
            assert database is not None  # For the linter's sake
            if isinstance(database, dict):
                # If a dictionary is provided, make sure all entries are correct
                self.database = {
                    run_id: self.load_value(val, num_tpcs)
                    for run_id, val in database.items()
                }
            else:
                self.database = CalibrationDatabase(database, num_tpcs)

    def load_value(self, value: Union[float, List[float]], num_tpcs: int) -> np.ndarray:
        """Process value whether it is provided as a scalar or a list.

        Parameters
        ----------
        value : Union[float, list]
            Specifies the quantity value as a scalar or a list
        num_tpcs : int
            Number of TPCs in the detector

        Returns
        -------
        np.ndarray
             List of quantities, one per TPC
        """
        # Inititalize quantity
        if not isinstance(value, (list, tuple, np.ndarray)):
            return np.full(num_tpcs, value)

        assert len(value) == num_tpcs, (
            "Calibration constant must be specified as either a scalar "
            f"or a list of length {num_tpcs}."
        )
        return np.asarray(value)

    def get(self, tpc_id: int, run_id: Optional[int] = None) -> float:
        """Fetch the calibration constant for a specific TPC and run.

        Parameters
        ----------
        tpc_id : int
            ID of the TPC to use
        run_id : int, optional
            ID of the run to get the value for, if using a database

        Returns
        -------
        float
            Calibration constant for the specified TPC (and run ID)
        """
        # Fetch the calibration constant
        if self.database is not None:
            assert run_id is not None, (
                "When using a calibration database, the run_id must be provided "
                "to fetch the correct gain value."
            )
            return self.database[run_id][tpc_id]

        elif self.value is not None:
            return self.value[tpc_id]

        else:
            raise RuntimeError("CalibrationConstant not properly initialized.")



"""Apply electron lifetime corrections."""

from typing import Dict, List, Optional, Union

import numpy as np

from spine.geo.base import Geometry

from .constant import CalibrationConstant

__all__ = ["LifetimeCalibrator"]


class LifetimeCalibrator:
    """Applies a correction based on drift electron lifetime and the distance
    from the ionization point to the closest readout plane.
    """

    name = "lifetime"

    def __init__(
        self,
        num_tpcs: int,
        lifetime: Optional[Union[float, List[float]]] = None,
        driftv: Optional[Union[float, List[float]]] = None,
        lifetime_db: Optional[Union[str, Dict[int, Union[float, List[float]]]]] = None,
        driftv_db: Optional[Union[str, Dict[int, Union[float, List[float]]]]] = None,
    ):
        """Load the information needed to make a lifetime correction.

        Parameters
        ----------
        num_tpcs : int
            Number of TPCs in the detector
        lifetime : Union[float, List[float]], optional
            Specifies the electron lifetime in microseconds. If `List`, it
            should map a tpc ID onto a specific value.
        driftv : Union[float, List[float]], optional
            Specifies the electron drift velocity in cm/us. If `List`, it
            should map a tpc ID onto a specific value.
        lifetime_db : Union[str, Dict[int, Union[float, List[float]]]], optional
            Path to a SQLite db file which maps [run, cryo, tpc] sets onto
            a specific lifetime value in microseconds, or a dictionary which
            maps run IDs to lifetime values.
        driftv_db : Union[str, Dict[int, Union[float, List[float]]]], optional
            Path to a SQLite db file which maps [run, cryo, tpc] sets onto
            a specific electron drift velocity value in cm/us, or a dictionary which
            maps run IDs to drift velocity values.
        """
        # Initialize lifetime and drift velocity calibration constants
        self.lifetime = CalibrationConstant(
            num_tpcs=num_tpcs, value=lifetime, database=lifetime_db
        )
        self.driftv = CalibrationConstant(
            num_tpcs=num_tpcs, value=driftv, database=driftv_db
        )

    def process(
        self,
        points: np.ndarray,
        values: np.ndarray,
        geo: Geometry,
        tpc_id: int,
        run_id: Optional[int] = None,
    ):
        """Apply the lifetime correction.

        Parameters
        ----------
        points : np.ndarray
            (N, 3) array of point coordinates
        values : np.ndarray
            (N) array of values associated with each point
        geo : Geometry
            Detector geometry object
        tpc_id : int
            ID of the TPC to use
        run_id : int, optional
            If provided, used to get the appropriate lifetime/drift velocities

        Returns
        -------
        np.ndarray
            (N) array of corrected values
        """
        # Get the corrections lifetimes/drift velocities
        lifetime = self.lifetime.get(tpc_id, run_id)
        driftv = self.driftv.get(tpc_id, run_id)

        # Compute the distance to the anode plane
        mod = tpc_id // geo.tpc.num_chambers_per_module
        tpc = tpc_id % geo.tpc.num_chambers_per_module
        daxis = geo.tpc[mod][tpc].drift_axis
        position = geo.tpc[mod][tpc].anode_pos
        drifts = np.abs(points[:, daxis] - position)

        # Clip down to the physical range of possible drift distances
        max_drift = geo.tpc[mod][tpc].dimensions[daxis]
        drifts = np.clip(drifts, 0.0, max_drift)

        # Convert the drift distances to correction factors
        corrections = np.exp(drifts / lifetime / driftv)
        return corrections * values



"""Track end point assignment module."""

from spine.post.base import PostBase
from spine.utils.globals import TRACK_SHP
from spine.utils.ppn import check_track_orientation_ppn
from spine.utils.tracking import check_track_orientation

__all__ = ["TrackExtremaProcessor"]


class TrackExtremaProcessor(PostBase):
    """Assigns track start point and end point."""

    # Name of the post-processor (as specified in the configuration)
    name = "track_extrema"

    # Alternative allowed names of the post-processor
    aliases = ("assign_track_extrema",)

    # Set of data keys needed for this post-processor to operate
    _keys = (("ppn_candidates", False),)

    def __init__(self, method="local", obj_type="particle", **kwargs):
        """Initialize the track end point assignment parameters.

        Parameters
        ----------
        method : str, default 'local'
            Algorithm to correct track startpoint/endpoint misplacement. The
            following methods are available:
            - local: computes local energy deposition density only at
            the extrema and chooses the higher one as the endpoint.
            - gradient: computes local energy deposition density throughout
            the track, computes the overall slope (linear fit) of the energy
            density variation to estimate the direction.
            - ppn: uses ppn candidate predictions (classify_endpoints) to
            assign start and endpoints.
        kwargs : dict
            Extra arguments to pass to the `check_track_orientation` or the
            `check_track_orientation_ppn' functions
        """
        # Initialize the parent class
        super().__init__(obj_type, "reco")

        # Store the orientation method and its arguments
        self.method = method
        self.kwargs = kwargs

    def process(self, data):
        """Assign track end points in one entry

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        for part in data["reco_particles"]:
            # Only process track objects
            if part.shape == TRACK_SHP:
                # Check if the end points need to be flipped
                if self.method in ["local", "gradient"]:
                    flip = not check_track_orientation(
                        part.points,
                        part.depositions,
                        part.start_point,
                        part.end_point,
                        self.method,
                        **self.kwargs,
                    )

                elif self.method == "ppn":
                    assert "ppn_candidates" in data, (
                        "Must run the `ppn` post-processor "
                        "before using PPN predictions to assign extrema."
                    )
                    flip = not check_track_orientation_ppn(
                        part.start_point, part.end_point, data["ppn_candidates"]
                    )

                else:
                    raise ValueError(
                        "Point assignment method not recognized: " f"{self.method}"
                    )

                # If needed, flip the end points and the directions
                if flip:
                    part.start_point, part.end_point = (
                        part.end_point,
                        part.start_point,
                    )
                    part.start_dir, part.end_dir = (-part.end_dir, -part.start_dir)



"""Multiple-Coulomb scattering (MCS) energy reconstruction module."""

import numpy as np

from spine.post.base import PostBase
from spine.utils.globals import (
    KAON_PID,
    MUON_PID,
    PID_MASSES,
    PION_PID,
    PROT_PID,
    TRACK_SHP,
)
from spine.utils.mcs import ANGLE_METHODS, mcs_angles, mcs_fit
from spine.utils.tracking import get_track_segments

__all__ = ["MCSEnergyProcessor"]


class MCSEnergyProcessor(PostBase):
    """Reconstruct the kinetic energy of tracks based on their Multiple-Coulomb
    scattering (MCS) angles while passing through liquid argon.
    """

    # Name of the post-processor (as specified in the configuration)
    name = "mcs_ke"

    # Alternative allowed names of the post-processor
    aliases = ("reconstruct_mcs_energy",)

    def __init__(
        self,
        tracking_mode="bin_pca",
        segment_length=5.0,
        res_a=0.25,
        res_b=1.25,
        res_mixture=False,
        res_weight_ratio=0.5,
        res_scale_ratio=2.25,
        include_pids=(MUON_PID, PION_PID, PROT_PID, KAON_PID),
        fill_per_pid=False,
        only_uncontained=False,
        angle_method="atan2",
        obj_type="particle",
        run_mode="both",
        truth_point_mode="points",
        pid_mode="pid",
        **kwargs,
    ):
        """Store the necessary attributes to do MCS-based estimations.

        Parameters
        ----------
        tracking_mode : str, default 'bin_pca'
            Method used to compute the segment angles (one of 'step',
            'step_next' or 'bin_pca')
        segment_length : float, default 5.0 cm
            Segment length in the units that specify the coordinates
        res_a : float, default 0.25 rad*cm^res_b
            Parameter a in the a/dx^b which models the angular uncertainty
        res_b : float, default 1.25
            Parameter b in the a/dx^b which models the angular uncertainty
        res_mixture : bool, default False
            If `True`, the detector resolution is modeled by a Rayleigh mixture
        res_weight_ratio : float, default 0.5
            When using a Rayleigh mixture, defines the weight ratio between components
        res_scale_ratio : float, default 2.25
            When using a Rayleigh mixture, defines the scale ratio between components
        include_pids : list, default [2, 3, 4, 5]
            Particle species to compute the kinetic energy for
        fill_per_pid : bool, default False
            If `True`, compute the MCS KE estimate under all PID assumptions
        only_uncontained : bool, default False
            Only run the algorithm on particles that are not contained
        angle_method : str, default 'atan2'
            Angular reconstruction method
        **kwargs : dict, optiona
            Additional arguments to pass to the tracking algorithm
        """
        # Initialize the parent class
        super().__init__(obj_type, run_mode, truth_point_mode, pid_mode=pid_mode)

        # Store the general parameters
        self.include_pids = include_pids
        self.fill_per_pid = fill_per_pid
        self.only_uncontained = only_uncontained

        # Store the tracking parameters
        assert tracking_mode in [
            "step",
            "step_next",
            "bin_pca",
        ], "The tracking algorithm must provide segment angles"
        self.tracking_mode = tracking_mode
        self.tracking_kwargs = kwargs

        # Store the angle computation parameters
        assert angle_method in ANGLE_METHODS, (
            f"Angular reconstruction method not recognized: {angle_method}. "
            f"Must be one of {ANGLE_METHODS.keys()}"
        )
        self.angle_method = ANGLE_METHODS[angle_method]

        # Store the MCS parameters
        self.segment_length = segment_length
        self.res_a = res_a
        self.res_b = res_b
        self.res_mixture = res_mixture
        self.res_weight_ratio = res_weight_ratio
        self.res_scale_ratio = res_scale_ratio

    def process(self, data):
        """Reconstruct the MCS KE estimates for each particle in one entry.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Loop over particle objects
        for k in self.fragment_keys + self.particle_keys:
            for obj in data[k]:
                # Only run this algorithm on particle species that are needed
                if not (
                    (obj.shape == TRACK_SHP)
                    and (self.get_pid(obj) in self.include_pids)
                ):
                    continue
                if self.only_uncontained and obj.is_contained:
                    continue

                # Make sure the particle coordinates are expressed in cm
                self.check_units(obj)

                # Get point coordinates
                points = self.get_points(obj)
                if len(points) == 0:
                    continue

                # Get the list of segment directions
                _, dirs, _ = get_track_segments(
                    points,
                    self.segment_length,
                    obj.start_point,
                    method=self.tracking_mode,
                    **self.tracking_kwargs,
                )

                # Find the angles between successive segments
                theta = mcs_angles(dirs, self.angle_method)
                if len(theta) < 1:
                    continue

                # Store the length and the MCS kinetic energy
                mass = PID_MASSES[self.get_pid(obj)]
                obj.mcs_ke = mcs_fit(
                    theta,
                    mass,
                    self.segment_length,
                    1,
                    self.res_a,
                    self.res_b,
                    self.res_mixture,
                    self.res_weight_ratio,
                    self.res_scale_ratio,
                )

                # If requested, convert the KE to other PID hypotheses
                if self.fill_per_pid:
                    # Compute the momentum (what MCS is truly sensitive to)
                    mom = np.sqrt(obj.mcs_ke**2 + 2 * mass * obj.mcs_ke)

                    # For each PID, convert back to KE
                    for pid in self.include_pids:
                        mass = PID_MASSES[pid]
                        ke = np.sqrt(mom**2 + mass**2) - mass
                        obj.mcs_ke_per_pid[pid] = ke



"""Feature embedding for pixel supervised connected-component clustering."""

import MinkowskiEngine as ME
import torch
import torch.nn as nn

from spine.data import TensorBatch
from spine.model.layer.cnn.uresnet_layers import UResNet
from spine.utils.globals import COORD_COLS, VALUE_COL

__all__ = ["GraphSPICEEmbedder"]


class GraphSPICEEmbedder(nn.Module):
    """Model which produces embeddings of an input sparse point cloud."""

    MODULES = ["uresnet"]

    def __init__(self, uresnet, **base):
        """Initialize the embedding model.

        Parameters
        ----------
        uresnet : dict
            Backbone UResNet configuration
        **base : dict, optional
            Basic parameters
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the uresnet backbone
        self.backbone = UResNet(uresnet)
        self.num_filters = self.backbone.num_filters
        self.spatial_size = self.backbone.spatial_size
        assert (
            self.spatial_size is not None
        ), "Must provide a spatial size to compute normalized coordinates."

        # Process the rest of the configuration
        self.process_model_config(**base)

        # Define output layers, if there is a need for them
        if not self.use_raw_features:
            self.out_spatial = nn.Sequential(
                nn.Linear(self.num_filters, self.spatial_embedding_dim), nn.Tanh()
            )
            self.out_feature = nn.Linear(self.num_filters, self.feature_embedding_dim)
            self.out_cov = nn.Linear(self.num_filters, 2)
            self.out_occupancy = nn.Linear(self.num_filters, 1)

        if self.predict_semantics:
            assert (
                self.num_classes is not None
            ), "Must specify the number of classes predicting semantics."
            self.out_seg = nn.Linear(self.num_filters, self.num_classes)

    def process_model_config(
        self,
        predict_semantics=False,
        num_classes=None,
        coord_conv=True,
        covariance_mode="softplus",
        occupancy_mode="softplus",
        feature_embedding_dim=16,
        spatial_embedding_dim=3,
        use_raw_features=False,
    ):
        """Process the embedding parameters.

        Parameters
        ----------
        predict_semantics : bool, default False
            If `True`, the embedder will output semantic predictions
        num_classes : int, optional
            Number of classes to classify the voxels as
        coord_conv : bool, default True
            If `True`, include the normalized pixel coordinates as a set of
            input features to the backbone UResNet
        covariance_mode : str, default 'softplus'
            Activation used to predict cluster covariance (spatial extent)
        occupancy_mode : str, default 'softplus'
            Activation used to predict cluster occupancy (pixel count)
        feature_embedding_dim : int, default 16
            Number of features per pixel in embedding space
        spatial_embedding_space : int, default 3
            Number of spatial features per pixel in embedding space
        use_raw_features : bool, default False
            Use the list of embedder features as is, without the output layers
        """
        # Store basic properties
        self.num_classes = num_classes
        self.coord_conv = coord_conv
        self.predict_semantics = predict_semantics
        self.use_raw_features = use_raw_features
        self.covariance_mode = covariance_mode
        self.occupancy_mode = occupancy_mode

        self.feature_embedding_dim = feature_embedding_dim
        self.spatial_embedding_dim = spatial_embedding_dim
        self.hyper_dimension = (
            self.spatial_embedding_dim + self.feature_embedding_dim + 3
        )

        # Initialize covariance activation function
        if self.covariance_mode == "exp":
            self.cov_func = torch.exp
        elif self.covariance_mode == "softplus":
            self.cov_func = nn.Softplus()
        else:
            raise ValueError(f"Covariance mode not recognized: {self.covariance_mode}")

        # Initialize occupancy activation function
        if self.occupancy_mode == "exp":
            self.occ_func = torch.exp
        elif self.occupancy_mode == "softplus":
            self.occ_func = nn.Softplus()
        else:
            raise ValueError(f"Occupancy mode not recognized: {self.covariance_mode}")

    def forward(self, data):
        """Compute the embeddings for one batch of data.

        Inputs
        ------
        data: TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
            - N is the the total number of voxels in the image
            - 1 is the batch ID
            - D is the number of dimensions in the input image
            - N_f is the number of features per voxel

        Returns
        -------
        dict
            Dictionary of outputs
        """
        # Build an input feature tensor
        coords = data.tensor[:, :VALUE_COL]
        features = data.tensor[:, VALUE_COL].view(-1, 1)

        # If requested, append the normalized coordinates to the feature tensor
        half_size = self.spatial_size / 2
        points = coords[:, 1:]
        normalized_coords = (points - half_size) / half_size
        if self.coord_conv:
            features = torch.cat([normalized_coords, features], dim=1)

        # Pass it through the backbone UResNet, extract output features
        backbone_data = torch.cat((coords, features), dim=1)
        result_backbone = self.backbone(backbone_data)
        output_features = result_backbone["decoder_tensors"][-1].F

        # Convert the output to tensor batches
        coords = TensorBatch(coords, data.counts, coord_cols=COORD_COLS)
        features = TensorBatch(output_features, data.counts)

        # Intialize the result
        result = {"coordinates": coords, "features": features}

        # If requested, pass the raw output features through final layers
        if not self.use_raw_features:
            # Spatial Embeddings (offset by the normalized coordinates)
            spatial_embeddings = self.out_spatial(output_features)

            # Feature Embeddings
            feature_embeddings = self.out_feature(output_features)

            # Covariance
            out = self.out_cov(output_features)
            covariance = self.cov_func(out)

            # Occupancy
            out = self.out_occupancy(output_features)
            occupancy = self.occ_func(out)

            # Bundle the features together
            hypergraph_features = torch.cat(
                [spatial_embeddings, feature_embeddings, covariance, occupancy], dim=1
            )

            # Convert the output to tensor batches
            spatial_embeddings = TensorBatch(
                spatial_embeddings + normalized_coords, data.counts
            )
            feature_embeddings = TensorBatch(feature_embeddings, data.counts)
            covariance = TensorBatch(covariance, data.counts)
            occupancy = TensorBatch(occupancy, data.counts)
            hypergraphneeded_features = TensorBatch(hypergraph_features, data.counts)

            # Append results
            result.update(
                {
                    "spatial_embeddings": spatial_embeddings,
                    "feature_embeddings": feature_embeddings,
                    "covariance": covariance,
                    "occupancy": occupancy,
                    "hypergraph_features": hypergraph_features,
                }
            )

        # If requested, add a semantic prediction to the output
        if self.predict_semantics:
            # Segmentation layer
            segmentation = self.out_seg(output_features)

            # Append results
            result["segmentation"] = TensorBatch(segmentation, data.counts)

        return result



"""UResNet segmentation model and its loss."""

from collections import defaultdict

import MinkowskiEngine as ME
import numpy as np
import torch
import torch.nn as nn

from spine.data import TensorBatch
from spine.utils.globals import BATCH_COL, COORD_COLS, GHOST_SHP, VALUE_COL
from spine.utils.logger import logger
from spine.utils.torch.scripts import cdist_fast

from .layer.cnn.act_norm import act_factory, norm_factory
from .layer.cnn.uresnet_layers import UResNet
from .layer.factories import loss_fn_factory

__all__ = ["UResNetSegmentation", "SegmentationLoss"]


class UResNetSegmentation(nn.Module):
    """UResNet for semantic segmentation.

    Typical configuration should look like:

    .. code-block:: yaml

        model:
          name: uresnet
          modules:
            uresnet:
              # Your config goes here

    See :func:`setup_cnn_configuration` for available parameters for the
    backbone UResNet architecture.

    See configuration file(s) prefixed with `uresnet_` under the `config`
    directory for detailed examples of working configurations.
    """

    INPUT_SCHEMA = [["sparse3d", (float,), (3, 1)]]

    MODULES = ["uresnet"]

    def __init__(self, uresnet, uresnet_loss=None):
        """Initializes the standalone UResNet model.

        Parameters
        ----------
        uresnet : dict
            Model configuration
        uresnet_loss : dict, optional
            Loss configuration
        """
        # Initialize the parent class
        super().__init__()

        # Initialize the model configuration
        self.process_model_config(**uresnet)

        # Initialize the output layer
        self.output = [
            norm_factory(self.backbone.norm_cfg, self.num_filters),
            act_factory(self.backbone.act_cfg),
        ]
        self.output = nn.Sequential(*self.output)
        self.linear_segmentation = ME.MinkowskiLinear(
            self.num_filters, self.num_classes
        )

        # If needed, activate the ghost classification layer
        if self.ghost:
            logger.debug("Ghost Masking is enabled for UResNet Segmentation")
            self.linear_ghost = ME.MinkowskiLinear(self.num_filters, 2)

    def process_model_config(self, num_classes, ghost=False, **backbone):
        """Initialize the underlying UResNet model.

        Parameters
        ----------
        num_classes : int
            Number of classes to classify the voxels as
        ghost : bool, default False
            Whether to add a deghosting step in the classification model
        **backbone : dict
            UResNet backbone configuration
        """
        # Store the semantic segmentation configuration
        self.num_classes = num_classes
        self.ghost = ghost

        # Initialize the UResNet backbone, store the relevant parameters
        self.backbone = UResNet(backbone)
        self.num_filters = self.backbone.num_filters

    def forward(self, data):
        """Run a batch of data through the forward function.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) tensor of voxel/value pairs
            - N is the the total number of voxels in the image
            - 1 is the batch ID
            - D is the number of dimensions in the input image
            - N_f is the number of features per voxel

        Returns
        -------
        dict
            Dictionary of outputs
        """
        # Restrict the input to the requested number of features
        num_cols = 1 + self.backbone.dim + self.backbone.num_input
        input_tensor = data.tensor[:, :num_cols]

        # Pass the data through the UResNet backbone
        result_backbone = self.backbone(input_tensor)

        # Pass the output features through the output layer
        feats = result_backbone["decoder_tensors"][-1]
        feats = self.output(feats)
        seg = self.linear_segmentation(feats)

        # Store the output as tensor batches
        segmentation = TensorBatch(seg.F, data.counts)

        batch_size = data.batch_size
        final_tensor = TensorBatch(
            result_backbone["final_tensor"], batch_size=batch_size, is_sparse=True
        )
        encoder_tensors = [
            TensorBatch(t, batch_size=batch_size, is_sparse=True)
            for t in result_backbone["encoder_tensors"]
        ]
        decoder_tensors = [
            TensorBatch(t, batch_size=batch_size, is_sparse=True)
            for t in result_backbone["decoder_tensors"]
        ]

        result = {
            "segmentation": segmentation,
            "final_tensor": final_tensor,
            "encoder_tensors": encoder_tensors,
            "decoder_tensors": decoder_tensors,
        }

        # If needed, pass the output features through the ghost linear layer
        if self.ghost:
            ghost = self.linear_ghost(feats)

            result["ghost"] = TensorBatch(ghost.F, data.counts)
            result["ghost_tensor"] = TensorBatch(ghost, data.counts, is_sparse=True)

        return result


class SegmentationLoss(torch.nn.modules.loss._Loss):
    """Loss definition for semantic segmentation.

    For a regular flavor UResNet, it is a cross-entropy loss.
    For deghosting, it depends on a configuration parameter `ghost`:

    - If `ghost=True`, we first compute the cross-entropy loss on the ghost
      point classification (weighted on the fly with sample statistics). Then we
      compute a mask = all non-ghost points (based on true information in label)
      and within this mask, compute a cross-entropy loss for the rest of classes.

    - If `ghost=False`, we compute a N+1-classes cross-entropy loss, where N is
      the number of classes, not counting the ghost point class.

    See Also
    --------
    :class:`UResNetSegmentation`
    """

    INPUT_SCHEMA = [["parse_sparse3d", (int,), (3, 1)]]

    def __init__(self, uresnet, uresnet_loss):
        """
        Initializes the segmentation loss

        Parameters
        ----------
        uresnet : dict
            Model configuration
        uresnet_loss : dict
            Loss configuration
        """
        # Initialize the parent class
        super().__init__()

        # Initialize what we need from the model configuration
        self.process_model_config(**uresnet)

        # Initialize the loss configuration
        self.process_loss_config(**uresnet_loss)

    def process_model_config(self, num_classes, ghost=False, **kwargs):
        """Process the parameters of the upstream model needed for in the loss.

        Parameters
        ----------
        num_classes : int
            Number of classes to classify the voxels as
        ghost : bool, default False
            Whether to add a deghosting step in the classification model
        **kwargs : dict, optional
            Leftover model configuration (no need in the loss)
        """
        # Store the semantic segmentation configuration
        self.num_classes = num_classes
        self.ghost = ghost

    def process_loss_config(
        self,
        loss="ce",
        ghost_label=-1,
        alpha=1.0,
        beta=1.0,
        balance_loss=False,
        upweight_points=False,
        upweight_radius=20,
    ):
        """Process the loss function parameters.

        Parameters
        ----------
        loss : str, default 'ce'
            Loss function used for semantic segmentation
        ghost_label : int, default -1
            ID of ghost points. If specified (> -1), classify ghosts only
        alpha : float, default 1.0
            Classification loss prefactor
        beta : float, default 1.0
            Ghost mask loss prefactor
        balance_loss : bool, default False
            Whether to weight the loss to account for class imbalance
        upweight_points : bool, default False
            Whether to weight the loss higher near specific points (to be
            provided as `point_label` as a loss input)
        upweight_radius: bool, default False
            Radius around the points of interest for which to upweight the loss
        """
        # Set the loss function
        self.loss_fn = loss_fn_factory(loss, reduction="none")

        # Store the loss configuration
        self.ghost_label = ghost_label
        self.alpha = alpha
        self.beta = beta
        self.balance_loss = balance_loss
        self.upweight_points = upweight_points
        self.upweight_radius = upweight_radius

        # If a ghost label is provided, it cannot be in conjecture with
        # having a dedicated ghost masking layer
        assert not (self.ghost and self.ghost_label > -1), (
            "Cannot classify ghost exclusively (ghost_label) and "
            "have a dedicated ghost masking layer at the same time."
        )

    def forward(
        self,
        seg_label,
        segmentation,
        point_label=None,
        ghost=None,
        weights=None,
        **kwargs,
    ):
        """Computes the cross-entropy loss of the semantic segmentation
        predictions.

        Parameters
        ----------
        seg_label : TensorBatch
            (N, 1 + D + 1) Tensor of segmentation labels for the batch
        segmentation : TensorBatch
            (N, N_c) Tensor of logits from the segmentation model
        point_label : TensorBatch, optional
            (P, 1 + D + 1) Tensor of points of interests for the batch. This
            is used to upweight the loss near specific points.
        ghost : TensorBatch, optional
            (N, 2) Tensor of ghost logits from the segmentation model
        weights : TensorBatch, optional
            (N) Tensor of weights for each pixel in the batch
        **kwargs : dict, optional
            Other outputs of the upstream model which are not relevant here

        Returns
        -------
        dict
            Dictionary of accuracies and losses
        """
        # Get the underlying tensor in each TensorBatch
        seg_label_t = seg_label.tensor
        segmentation_t = segmentation.tensor
        ghost_t = ghost.tensor if ghost is not None else ghost
        weights_t = weights.tensor if weights is not None else weights

        # Make sure that the segmentation output and labels have the same length
        assert len(seg_label_t) == len(segmentation_t), (
            f"The `segmentation` output length ({len(segmentation_t)}) "
            f"and its labels ({len(seg_label_t)}) do not match."
        )
        assert not self.ghost or len(seg_label_t) == len(ghost_t), (
            f"The `ghost` output length ({len(ghost_t)}) and "
            f"its labels ({len(seg_label_t)}) do not match."
        )
        assert not self.ghost or weights is None, (
            "Providing explicit weights is not compatible when peforming "
            "deghosting in tandem with semantic segmentation."
        )

        # If requested, produce weights based on point-proximity
        if self.upweight_points:
            assert point_label is not None, (
                "If upweighting the loss nearby points of interests, must "
                "provide a list of such points in `point_label`."
            )
            dist_weights = self.get_distance_weights(seg_label, point_label)
            if weights is not None:
                weights_t *= dist_weights.tensor
            else:
                weights_t = dist_weights

        # Check that the labels have sensible values
        if self.ghost_label > -1:
            labels_t = (seg_label_t[:, VALUE_COL] == self.ghost_label).long()

        else:
            labels_t = seg_label_t[:, VALUE_COL].long()
            if torch.any(labels_t > self.num_classes):
                raise ValueError(
                    "The segmentation labels contain labels larger than "
                    "the number of logits output by the model."
                )

        # If there is a dedicated ghost layer, apply the mask first
        if self.ghost:
            # Count the number of voxels in each class
            ghost_labels_t = (labels_t == GHOST_SHP).long()
            ghost_loss, ghost_acc, ghost_acc_class = self.get_loss_accuracy(
                ghost_t, ghost_labels_t
            )

            # Restrict the segmentation target to true non-ghosts
            nonghost = torch.nonzero(ghost_labels_t == 0).flatten()
            segmentation_t = segmentation_t[nonghost]
            labels_t = labels_t[nonghost]

        # Compute the loss/accuracy of the semantic segmentation step
        seg_loss, seg_acc, seg_acc_class, weights_t = self.get_loss_accuracy(
            segmentation_t, labels_t, weights_t
        )

        # Get the combined loss and accuracies
        result = {}
        if self.ghost:
            result.update(
                {
                    "loss": self.alpha * seg_loss + self.beta * ghost_loss,
                    "accuracy": (seg_acc + ghost_acc) / 2.0,
                    "seg_loss": seg_loss,
                    "seg_accuracy": seg_acc,
                    "ghost_loss": ghost_loss,
                    "ghost_accuracy": ghost_acc,
                    "ghost_accuracy_class_0": ghost_acc_class[0],
                    "ghost_accuracy_class_1": ghost_acc_class[1],
                }
            )

            for c in range(self.num_classes):
                result[f"seg_accuracy_class_{c}"] = seg_acc_class[c]

        else:
            result.update({"loss": seg_loss, "accuracy": seg_acc})

            for c in range(self.num_classes):
                result[f"accuracy_class_{c}"] = seg_acc_class[c]

        if weights_t is not None:
            result["weights"] = TensorBatch(weights_t, seg_label.counts)

        return result

    def get_distance_weights(self, seg_label, point_label):
        """Define weights for each of the points in the image based on their
        distance from points of interests (typically vertices, but user defined).

        Parameters
        ----------
        seg_label : TensorBatch
            (N, 1 + D + 1) Tensor of segmentation labels for the batch
        point_label : TensorBatch
            (P, 1 + D + 1) Tensor of points of interests for the batch. This
            is used to upweight the loss of points near a vertex.

        Returns
        -------
        torch.Tensor
            (N) Array of weights associated with each point
        """
        # Loop over the entries in the batch, compute proximity for each point
        dists = torch.full_like(seg_label.tensor[:, 0], float("inf"))
        for b in range(seg_label.batch_size):
            # Fetch image voxel and point coordinates for this entry
            voxels_b = seg_label[b][:, COORD_COLS]
            points_b = point_label[b][:, COORD_COLS]
            if not len(points_b) or not len(voxels_b):
                continue

            # Compute the minimal distance to any point in this entry
            dist_mat = cdist_fast(voxels_b, points_b)
            dists_b = torch.min(dist_mat, dim=1).values

            # Record information in the batch-wise tensor
            lower, upper = seg_label.edges[b], seg_label.edges[b + 1]
            dists[lower:upper] = dists_b

        # Upweight the points within some distance of the points of interest
        proximity = (dists < self.upweight_radius).long()
        close_count = torch.sum(proximity)
        counts = torch.tensor(
            [len(dists) - close_count, close_count],
            dtype=torch.long,
            device=dists.device,
        )
        weights = len(proximity) / 2 / counts

        return weights[proximity]

    def get_loss_accuracy(self, logits, labels, weights=None):
        """Computes the loss, global and classwise accuracy.

        Parameters
        ----------
        logits : torch.Tensor
            (N, N_c) Output logits from the network for each voxel
        labels : torch.Tensor
            (N) Target values for each voxel
        weights : torch.Tensor, optional
            (N) Tensor of weights for each pixel in the batch

        Returns
        -------
        torch.Tensor
            Cross-entropy loss value
        float
            Global accuracy
        np.ndarray
            (N_c) Vector of class-wise accuracy
        torch.Tensor
            (N) Updated set of weights for each pixel in the batch
        """
        # If there is no input, nothing to do
        num_classes = logits.shape[1]
        if not len(logits):
            return 0.0, 1.0, np.ones(num_classes, dtype=np.float32), weights

        # Count the number of voxels in each class
        counts = torch.empty(num_classes, dtype=torch.long, device=labels.device)
        for c in range(num_classes):
            counts[c] = torch.sum(labels == c).item()

        # If requested, create a set of weights based on class prevalance
        if self.balance_loss:
            class_weight = torch.ones(
                len(counts), dtype=logits.dtype, device=logits.device
            )
            class_weight[counts > 0] = len(labels) / num_classes / counts[counts > 0]
            class_weights = class_weight[labels]
            if weights is not None:
                weights *= class_weights
            else:
                weights = class_weights

        # Compute the loss
        if weights is None:
            loss = self.loss_fn(logits, labels).mean()
        else:
            loss = (weights * self.loss_fn(logits, labels)).sum() / weights.sum()

        # Compute the accuracies
        with torch.no_grad():
            # Per-class prediction accuracy
            preds = torch.argmax(logits, dim=-1)
            acc_class = np.ones(num_classes, dtype=np.float32)
            for c in range(num_classes):
                if counts[c] > 0:
                    mask = torch.nonzero(labels == c).flatten()
                    acc_class[c] = (preds[mask] == c).sum().item() / counts[c]

            # Global prediction accuracy
            acc = (preds == labels).sum().item() / torch.sum(counts).item()

        return loss, acc, acc_class, weights



"""Module with a dataclass targeted at a batched edge index.

An edge index is a sparse representation of a graph incidence matrix.
"""

from dataclasses import dataclass
from typing import Union

import numpy as np

from spine.utils.conditional import torch
from spine.utils.docstring import inherit_docstring

from .base import BatchBase

__all__ = ["EdgeIndexBatch"]


@dataclass(eq=False)
@inherit_docstring(BatchBase)
class EdgeIndexBatch(BatchBase):
    """Batched edge index with the necessary methods to slice it.

    Attributes
    ----------
    offsets : Union[np.ndarray, torch.Tensor]
        (B) Offsets between successive indexes in the batch
    directed : bool
        Whether the edge index is directed or undirected
    """

    offsets: Union[np.ndarray, torch.Tensor]
    directed: bool

    def __init__(self, data, counts, offsets, directed):
        """Initialize the attributes of the class.

        If the edge index corresponds to an undirected graph, each edge
        should have its reciprocal edge immediately after, e.g.

        .. code-block:: python

            [[0,1,0,2,0,3,...],
             [1,0,2,0,3,0,...]]

        Parameters
        ----------
        data : Union[np.ndarray, torch.Tensor]
            (2, E) Batched edge index
        counts : Union[List[int], np.ndarray, torch.Tensor]
            (B) Number of index elements per entry in the batch
        offsets : Union[List[int], np.ndarray, torch.Tensor]
            (B) Offsets between successive indexes in the batch
        directed : bool
            Whether the edge index is directed or undirected
        """
        # Initialize the base class
        super().__init__(data)

        # Cast
        counts = self._as_long(counts)
        offsets = self._as_long(offsets)

        # Do a couple of basic sanity checks
        assert (
            self._sum(counts) == data.shape[1]
        ), "The `counts` provided do not add up to the index length"
        assert len(counts) == len(
            offsets
        ), "The number of `offsets` es not match the number of `counts`"
        if not directed:
            assert data.shape[1] % 2 == 0, (
                "If the edge index is undirected, it should have an "
                "even number of edge"
            )

        # Get the boundaries between successive index using the counts
        edges = self.get_edges(counts)

        # Store the attributes
        self.data = data
        self.counts = counts
        self.edges = edges
        self.offsets = offsets
        self.directed = directed
        self.batch_size = len(counts)

    def __getitem__(self, batch_id):
        """Returns a subset of the index corresponding to one entry.

        Parameters
        ----------
        batch_id : int
            Entry index
        """
        # Make sure the batch_id is sensible
        if batch_id >= self.batch_size:
            raise IndexError(
                f"Index {batch_id} out of bound for a batch size "
                f"of ({self.batch_size})"
            )

        # Return
        lower, upper = self.edges[batch_id], self.edges[batch_id + 1]
        index = self.data[:, lower:upper] - self.offsets[batch_id]
        return self._transpose(index)

    @property
    def index(self):
        """Alias for the underlying data stored.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (2, E) Underlying batch of edge indexes
        """
        return self.data

    @property
    def index_t(self):
        """Alias for the underlying data stored, transposed

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (E, 2) Underlying batch of edge indexes, transposed
        """
        return self._transpose(self.data)

    @property
    def batch_ids(self):
        """Returns the batch ID of each element in the full index list.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (N) Complete batch ID array, one per element
        """
        return self._repeat(self._arange(self.batch_size), self.counts)

    @property
    def directed_index(self):
        """Index of the directed graph. If a graph is undirected, it only
        returns one of the two edges corresponding to a connection.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (2, E//2) Underlying batch of edge indexes
        """
        # If the graph is directed, nothing to do
        if self.directed:
            return self.data

        # Otherwise, skip every second edge in the index
        return self.data[:, ::2]

    @property
    def directed_index_t(self):
        """Index of the directed graph, transposed. If the graph is undirected,
        it only returns one of the two edges corresponding to a connection.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (E//2, 2) Underlying batch of edge indexes, transposed
        """
        return self._transpose(self.directed_index)

    @property
    def directed_counts(self):
        """Returns the number of edges per entry, counting edges once even
        if they are bidirectional.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (B) Complete batch ID array, one per element
        """
        # If the graph is directed, the counts are exact
        if self.directed:
            return self.counts

        # Otherwise, indexes are twice as long
        return self.counts // 2

    @property
    def directed_batch_ids(self):
        """Returns the batch ID of each element in the directed index.

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (N) Complete batch ID array, one per element
        """
        return self._repeat(self._arange(self.batch_size), self.directed_counts)

    def split(self):
        """Breaks up the index batch into its constituents.

        Returns
        -------
        List[Union[np.ndarray, torch.Tensor]]
            List of one index per entry in the batch
        """
        indexes = self._split(self._transpose(self.index), self.splits)
        for batch_id in range(self.batch_size):
            indexes[batch_id] = indexes[batch_id] - self.offsets[batch_id]

        return indexes

    def to_numpy(self):
        """Cast underlying index to a `np.ndarray` and return a new instance.

        Returns
        -------
        TensorBatch
            New `TensorBatch` object with an underlying np.ndarray tensor.
        """
        # If the underlying data is of the right type, nothing to do
        if self.is_numpy:
            return self

        data = self._to_numpy(self.data)
        counts = self._to_numpy(self.counts)
        offsets = self._to_numpy(self.offsets)

        return EdgeIndexBatch(data, counts, offsets, self.directed)

    def to_tensor(self, dtype=None, device=None):
        """Cast underlying index to a `torch.tensor` and return a new instance.

        Parameters
        ----------
        dtype : torch.dtype, optional
            Data type of the tensor to create
        device : torch.device, optional
            Device on which to put the tensor

        Returns
        -------
        TensorBatch
            New `TensorBatch` object with an underlying np.ndarray tensor.
        """
        # If the underlying data is of the right type, nothing to do
        if not self.is_numpy:
            return self

        data = self._to_tensor(self.data, dtype, device)
        counts = self._to_tensor(self.counts, dtype, device)
        offsets = self._to_tensor(self.offsets, dtype, device)

        return EdgeIndexBatch(data, counts, offsets, self.directed)



import numba as nb
import numpy as np
import scipy

import spine.math as sm

from .energy_loss import step_energy_loss_lar
from .globals import LAR_X0

ANGLE_METHODS = {"cos": 0, "atan2": 1, "kahan": 2}


def mcs_fit(
    theta,
    M,
    dx,
    z=1,
    res_a=0.25,
    res_b=1.25,
    res_mixture=False,
    res_weight_ratio=0.5,
    res_scale_ratio=2.25,
    lower_bound=10.0,
    upper_bound=100000.0,
):
    """Finds the kinetic energy which best fits a set of scattering angles
    measured between successive segments along a particle track.

    Parameters
    ----------
    theta : np.ndarray
        (N) Vector of scattering angle at each step in radians
    M : float
        Particle mass in MeV/c^2
    dx : float
        Step length in cm
    z : int, default 1
        Impinging partile charge in multiples of electron charge
    res_a : float, default 0.25 rad*cm^res_b
        Parameter a in the a/dx^b which models the angular uncertainty
    res_b : float, default 1.25
        Parameter b in the a/dx^b which models the angular uncertainty
    res_mixture : bool, default False
        If `True`, a Rayleigh mixture is used to describe the angular resolution.
        Otherwise, a simple Rayleigh distribution is used
    res_weight_ratio : float, default 0.5
        When using a Rayleigh mixture, defines the weight ratio between components
    res_scale_ratio : float, default 2.25
        When using a Rayleigh mixture, defines the scale ratio between components
    lower_bound : float, default 10.
        Minimum allowed kinetic energy in MeV
    upper_bound : float, default 100000.
        Maximum allowed kinetic energy in MeV
    """
    # Optimize the initial kinetic energy given a set of angles
    fit_min = scipy.optimize.minimize_scalar(
        mcs_nll_lar,
        args=(
            theta,
            M,
            dx,
            z,
            res_a,
            res_b,
            res_mixture,
            res_weight_ratio,
            res_scale_ratio,
        ),
        bounds=[lower_bound, upper_bound],
    )

    return fit_min.x


@nb.njit(cache=True)
def mcs_nll_lar(
    T0,
    theta,
    M,
    dx,
    z=1,
    res_a=0.25,
    res_b=1.25,
    res_mixture=False,
    res_weight_ratio=0.5,
    res_scale_ratio=2.25,
):
    """Computes the MCS negative log likelihood for a given list of segment
    angles and an initial momentum. This function checks the agreement between
    the scattering expection and the observed scattering at each step.

    Parameters
    ----------
    T0 : float
        Candidate particle kinetic energy in MeV
    theta : np.ndarray
        (N) Vector of scattering angle at each step in radians
    M : float
        Particle mass in MeV/c^2
    dx : float
        Step length in cm
    z : int, default 1
       Impinging partile charge in multiples of electron charge
    res_a : float, default 0.25 rad*cm^res_b
        Parameter a in the a/dx^b which models the angular uncertainty
    res_b : float, default 1.25
        Parameter b in the a/dx^b which models the angular uncertainty
    res_mixture : bool, default False
        If `True`, a Rayleigh mixture is used to describe the angular resolution.
        Otherwise, a simple Rayleigh distribution is used
    res_weight_ratio : float, default 0.5
        When using a Rayleigh mixture, defines the weight ratio between components
    res_scale_ratio : float, default 2.25
        When using a Rayleigh mixture, defines the scale ratio between components
    """
    # Compute the kinetic energy at each step
    assert len(theta), "Must provide angles to esimate the MCS loss"
    num_steps = len(theta + 1)
    ke_array = step_energy_loss_lar(T0, M, dx, num_steps=num_steps)

    # If there are less steps than expected, T0 is too low
    if len(ke_array) < num_steps + 1:
        return np.inf

    # Convert the kinetic energy array to momenta
    mom_array = np.sqrt(ke_array**2 + 2 * M * ke_array)

    # Define each segment momentum as the geometric mean of its end points
    mom_steps = np.sqrt(mom_array[1:] * mom_array[:-1])

    # Get the expected scattering angle for each step
    theta0 = highland(mom_steps, M, dx, z)

    # Update the expected scattering angle to account for angular resolution
    res = res_a / dx**res_b

    # Compute the negative log likelihood
    if not res_mixture:
        # Add the resolution in quadrature
        theta0 = np.sqrt(theta0**2 + res**2)
        nll = np.sum(0.5 * (theta / theta0) ** 2 + 2 * np.log(theta0))

    else:
        # Add a mixture of resolutions
        res1, res2 = res, res * res_scale_ratio
        s1 = np.sqrt(theta0**2 + np.asarray(res1) ** 2)
        s2 = np.sqrt(theta0**2 + np.asarray(res2) ** 2)
        a = -2 * np.log(s1) - theta**2 / (2 * s1**2)
        b = -2 * np.log(s2) - theta**2 / (2 * s2**2)
        ll = np.logaddexp(a, np.log(res_weight_ratio) + b)
        nll = -np.sum(ll)

    return nll


@nb.njit(cache=True)
def highland(p, M, dx, z=1, X0=LAR_X0):
    """Highland scattering formula.

    Parameters
    ----------
    p : float
       Momentum in MeV/c
    M : float
       Impinging particle mass in MeV/c^2
    dx : float
        Step length in cm
    z : int, default 1
       Impinging partile charge in multiples of electron charge
    X0 : float, default LAR_X0
       Radiation length in the material of interest in cm

    Returns
    -------
    float
        Expected scattering angle in radians
    """
    # Highland formula
    beta = p / np.sqrt(p**2 + M**2)
    prefactor = 13.6 * z / beta / p

    return prefactor * (
        np.sqrt(dx / LAR_X0) * (1.0 + 0.038 * np.log(z**2 * dx / LAR_X0 / beta**2))
    )


@nb.njit(cache=True)
def mcs_angles(dirs, method=ANGLE_METHODS["atan2"], axis=None):
    """Compute successive angles based on a list of directions of segments.

    Parameters
    ----------
    dirs : np.ndarray
        (N, 3) Array of segments directions
    method : int, default 1
        Enumerated angle computation method
    axis : int, optional
        Axis with respect to which to orient a projected angle

    Returns
    -------
    np.ndarray
        (N - 1) Array of successive angles
    """
    if method == 0:
        return angles_cos(dirs, axis)
    elif method == 1:
        return angles_atan2(dirs, axis)
    elif method == 2:
        return angles_kahan(dirs, axis)
    else:
        raise ValueError("Angle measurement method not recognized")


@nb.njit(cache=True)
def mcs_angles_proj(dirs, method=ANGLE_METHODS["atan2"]):
    """Compute successive projected angles based on a list of directions of
    segments.

    Parameters
    ----------
    dirs : np.ndarray
        (N, 3) Array of segments directions
    method : int, default 1
        Enumerated angle computation method

    Returns
    -------
    np.ndarray
        (N - 1, 3) Array of successive angles (one per projection)
    """
    # Loop over axes to zero-out
    angles_proj = np.empty((len(dirs) - 1, 3), dtype=dirs.dtype)
    for axis in range(3):
        # Zero-out axis, normalize
        dirs_proj = np.copy(dirs)
        dirs_proj[:, axis] = 0.0
        dirs_proj /= sm.linalg.norm(dirs_proj, axis=1)[:, None]

        # Compute angles, provide a reference axis to return a signed angle
        angles_proj[:, axis] = mcs_angles(dirs_proj, method, axis=axis)

    return angles_proj


@nb.njit(cache=True)
def angles_cos(dirs, axis=None):
    """Compute successive angles based on a list of directions of segments using
    the simple cosine method (unstable near 0).

    Parameters
    ----------
    dirs : np.ndarray
        (N, 3) Array of segments directions
    axis : int, optional
        Axis with respect to which to orient a projected angle

    Returns
    -------
    np.ndarray
        (N - 1) Array of successive angles
    """
    cos = sm.sum(dirs[:-1] * dirs[1:], axis=1)
    angles = np.arccos(np.clip(cos, -1.0, 1.0))
    if axis is not None:
        signs = np.sign(np.cross(dirs[:-1], dirs[1:])[:, axis])
        angles *= signs

    return angles


@nb.njit(cache=True)
def angles_atan2(dirs, axis=None):
    """Compute successive angles based on a list of directions of segments using
    the atan2 method (with dot and cross products).

    Parameters
    ----------
    dirs : np.ndarray
        (N, 3) Array of segments directions
    axis : int, optional
        Axis with respect to which to orient a projected angle

    Returns
    -------
    np.ndarray
        (N - 1) Array of successive angles
    """
    cross = np.cross(dirs[:-1], dirs[1:])
    if axis is None:
        cross = sm.linalg.norm(cross, axis=1)
    else:
        cross = cross[:, axis]

    dot = sm.sum(dirs[:-1] * dirs[1:], axis=1)

    return np.arctan2(cross, dot)


@nb.njit(cache=True)
def angles_kahan(dirs, axis):
    """Compute successive angles based on a list of directions of segments using
    Kahan's half-angle trick.

    Parameters
    ----------
    dirs : np.ndarray
        (N, 3) Array of segments directions
    axis : int, optional
        Axis with respect to which to orient a projected angle

    Returns
    -------
    np.ndarray
        (N - 1) Array of successive angles
    """
    diff = sm.linalg.norm(dirs[1:] - dirs[:-1], axis=1)
    summ = sm.linalg.norm(dirs[1:] + dirs[:-1], axis=1)
    angles = 2 * np.arctan(diff / summ)
    if axis is not None:
        signs = np.sign(np.cross(dirs[:-1], dirs[1:])[:, axis])
        angles *= signs

    return angles


@nb.njit(cache=True)
def split_angles(theta):
    """Split 3D angles between vectors to two 2D projected angles.

    Parameters
    ----------
    theta : np.ndarray
        Array of 3D angles in [0, np.pi]

    Returns
    -------
    float
        Frist array of 2D angle in [-np.pi, np.pi]
    float
        Second array of 2D angle in [-np.pi, np.pi]
    """
    # Randomize the azimuthal angle
    phi = np.random.uniform(0, 2 * np.pi, len(theta))
    offset = (theta > np.pi / 2) * np.pi
    sign = (-1) ** (theta > np.pi / 2)

    # Project the 3D angle along the two planes
    radius = np.sin(theta)
    theta1 = offset + sign * np.arcsin(np.sin(phi) * radius)
    theta2 = offset + sign * np.arcsin(np.cos(phi) * radius)

    # If theta is over pi/2, must flip the angles
    return theta1, theta2



"""Connected component clustering module."""

import numpy as np
import torch

from spine.data import TensorBatch
from spine.math.graph import connected_components

from .orphan import OrphanAssigner

__all__ = ["ConnectedComponentClusterer"]


class ConnectedComponentClusterer:
    """Finds connected components based on graph edge assignments."""

    def __init__(self, min_size=0, orphan=None):
        """Initialize the connected component constructor.

        Parameters
        ----------
        min_size : int, default 0
            Minimum number of points in a label cluster to be considered
            already assigned. If below this value, it is considered an orphan.
        orphan : dict
            Orphan assigner configuration dictionary
        """
        # Store the parameters
        self.min_size = min_size

        # Initialize the orphan assigner
        self.orphan_assigner = None
        if orphan is not None:
            self.orphan_assigner = OrphanAssigner(**orphan)

    def __call__(
        self,
        node_coords,
        edge_index,
        edge_assn,
        node_clusts,
        edge_clusts,
        min_size=None,
    ):
        """Loop over the list of batch entries and semantic classes and
        label points use connected-component clustering.

        Parameters
        ----------
        node_coords : TensorBatch
            (N, 3) Set of point coordinates
        edge_index : TensorBatch
            (E, 2) Set of edge source and target indices
        edge_assn : TensorBatch
            (E) Boolean assignment for each edge (0 for off, 1 for on)
        node_clusts : List[List[List[int]]]
            (B, S) One list of node indices per (batch ID, shape) pair
        edge_clusts : List[List[List[int]]]
            (B, S) One list of edge indices per (batch ID, shape) pair
        min_size : int, optional
            Override the minimum cluster size set in the configuration

        Returns
        -------
        TensorBatch
            (N) Cluster assignments for each of the points in the input
        """
        # Cast the input to numpy, if necessary
        tensor_input = isinstance(node_coords.tensor, torch.Tensor)
        if tensor_input:
            device = node_coords.device
            node_clusts = node_clusts.to_numpy()
            edge_clusts = edge_clusts.to_numpy()
            node_coords = node_coords.to_numpy()
            edge_index = edge_index.to_numpy()
            edge_assn = edge_assn.to_numpy()

        # Loop over the unique entries in the batch
        node_pred_list = []
        for b in range(node_coords.batch_size):
            # Make predictions for one entry
            node_pred_b = self.fit_predict_entry(
                node_coords[b],
                edge_index[b],
                edge_assn[b],
                node_clusts[b],
                edge_clusts[b],
                min_size,
            )

            node_pred_list.append(node_pred_b)

        # Aggregate the node assignments into a batched object. If the input
        # is a torch tensor, convert the output
        node_pred = np.concatenate(node_pred_list)
        node_pred = TensorBatch(node_pred, counts=node_coords.counts)
        if tensor_input:
            node_pred = node_pred.to_tensor(dtype=torch.long, device=device)

        return node_pred

    def fit_predict_entry(
        self,
        node_coords,
        edge_index,
        edge_assn,
        node_clusts,
        edge_clusts,
        min_size=None,
    ):
        """Assign cluster labels to graph nodes based on edge assignments
        in one entry.

        Parameters
        ----------
        node_coords : np.ndarray
            (N, 3) Set of point coordinates
        edge_index : np.ndarray
            (E, 2) Set of edge source and target indices
        edge_assn : np.ndarray
            (E) Boolean assignment for each edge (0 for off, 1 for on)
        node_clusts : List[List[int]]
            (B, S) One list of node indices per (batch ID, shape) pair
        edge_clusts : List[List[int]]
            (B, S) One list of edge indices per (batch ID, shape) pair
        min_size : int, optional
            Override the minimum cluster size set in the configuration

        Returns
        -------
        np.ndarry
            (N) Cluster assignments for each of the points in the input
        """
        # Narrow down the input to this specific entry
        assert len(node_clusts) == len(edge_clusts), (
            "There should be one index per semantic class for both "
            "nodes and edges. Got different size arrays."
        )

        # Loop over the semantic classes
        offset = 0
        node_pred = -np.ones(len(node_coords), dtype=int)
        for nindex, eindex in zip(node_clusts, edge_clusts):
            # Get predictions for a specifc (batch ID, shape) pair
            node_pred_s = self.fit_predict_one(
                node_coords[nindex],
                edge_index[eindex],
                edge_assn[eindex],
                offset,
                min_size,
            )

            # Update the list of node assignments, offset the cluster
            # counter to make sure there is no overlap between shapes
            if len(nindex):
                node_pred[nindex] = node_pred_s
                offset = int(node_pred.max()) + 1

        return node_pred

    def fit_predict_one(
        self, node_coords, edge_index, edge_assn, offset, min_size=None
    ):
        """Assign cluster labels to graph nodes based on edge assignments
        in one graph, i.e. one (entry, shape) pair.

        Parameters
        ----------
        node_coords : np.ndarray
            (N, 3) Set of point coordinates
        edge_index : np.ndarray
            (E, 2) Set of edge source and target indices
        edge_assn : np.ndarray
            (E) Boolean assignment for each edge (0 for off, 1 for on)
        offset : int
            Offset to apply to assigned nodes
        min_size : int, optional
            Override the minimum cluster size set in the configuration

        Returns
        -------
        np.ndarry
            (N) Cluster assignments for each of the points in the input
        """
        # Narrow down the list of edges to those turned on
        assert edge_index.shape[1] == 2, "The edge index must be of shape (E, 2)."
        edges = edge_index[np.where(edge_assn)[0]]

        # Find connected components
        node_pred = connected_components(edges, len(node_coords))

        # If min_size is set, downselect the points considered labeled
        min_size = min_size if min_size is not None else self.min_size
        if min_size > 1:
            _, inverse, counts = np.unique(
                node_pred, return_inverse=True, return_counts=True
            )
            orphan_mask = np.where(counts[inverse] < min_size)[0]
            node_pred[orphan_mask] = -1

        # If requested, assign the orphans
        if self.orphan_assigner is not None:
            node_pred = self.orphan_assigner(node_coords, node_pred)

        # Offset the valid assignments
        node_pred[node_pred != -1] += offset

        return node_pred



"""Class which adapts clustering labels given upstream semantic predictions."""

import numpy as np
import torch

from spine.data import TensorBatch
from spine.math.distance import METRICS, cdist, get_metric_id
from spine.utils.globals import (
    CLUST_COL,
    COORD_COLS,
    DELTA_SHP,
    GHOST_SHP,
    MICHL_SHP,
    SHAPE_COL,
    SHOWR_SHP,
    TRACK_SHP,
    VALUE_COL,
)
from spine.utils.gnn.cluster import break_clusters
from spine.utils.torch.scripts import cdist_fast

__all__ = ["ClusterLabelAdapter"]


class ClusterLabelAdapter:
    """Adapts the cluster labels to account for the predicted semantics.

    Points wrongly predicted get the cluster label of the closest touching
    compatible cluster, if there is one. Points that are predicted as ghosts
    get invalid (-1) cluster labels everywhere.

    Instances that have been broken up by the deghosting or semantic
    segmentation process get assigned distinct cluster labels for each
    effective fragment, provided they appear in the `break_classes` list.

    Notes
    -----
    This class supports both Numpy arrays and Torch tensors.
    """

    def __init__(
        self,
        break_eps=1.1,
        break_metric="chebyshev",
        break_p=2.0,
        break_classes=[SHOWR_SHP, TRACK_SHP, MICHL_SHP, DELTA_SHP],
    ):
        """Initialize the adapter class.

        Parameters
        ----------
        dtype : str, default 'torch'
            Type of data to be processed through the label adapter
        break_eps : float, default 1.1
            Distance scale used in the break up procedure
        break_metric : str, default 'chebyshev'
            Distance metric used in the break up produce
        p : float, default 2.
            p-norm factor for the Minkowski metric, if used
        break_classes : List[int], default
                        [SHOWR_SHP, TRACK_SHP, MICHL_SHP, DELTA_SHP]
            Classes to run DBSCAN on to break up
        """
        # Store relevant parameters
        self.break_eps = break_eps
        self.break_metric_id = get_metric_id(break_metric, break_p)
        self.break_p = break_p
        self.break_classes = break_classes

        # Attributes used to fetch the correct functions
        self.torch, self.dtype, self.device = None, None, None

    def __call__(self, clust_label, seg_label, seg_pred, ghost_pred=None):
        """Adapts the cluster labels for one entry or a batch of entries.

        Parameters
        ----------
        clust_label : Union[TensorBatch, np.ndarray, torch.Tensor]
            (N, N_l) Cluster label tensor
        seg_label : Union[TensorBatch, np.ndarray, torch.Tensor]
            (M, 5) Segmentation label tensor
        seg_pred : Union[TensorBatch, np.ndarray, torch.Tensor]
            (M/N_deghost) Segmentation predictions for each voxel
        ghost_pred : Union[TensorBatch, np.ndarray, torch.Tensor], optional
            (M) Ghost predictions for each voxel

        Returns
        -------
        Union[TensorBatch, np.ndarray, torch.Tensor]
            (N_deghost, N_l) Adapted cluster label tensor
        """
        # Set the data type/device based on the input
        ref_tensor = clust_label
        if isinstance(ref_tensor, TensorBatch):
            ref_tensor = ref_tensor.tensor
        self.torch = isinstance(ref_tensor, torch.Tensor)

        self.dtype = clust_label.dtype
        if self.torch:
            self.device = clust_label.device

        # Dispatch depending on the data type
        self._offset = 0
        if isinstance(clust_label, TensorBatch):
            # If it is batch data, call the main process function of each entry
            shape = (seg_pred.shape[0], clust_label.shape[1])
            clust_label_adapted = torch.empty(
                shape, dtype=clust_label.dtype, device=clust_label.device
            )
            for b in range(clust_label.batch_size):
                lower, upper = seg_pred.edges[b], seg_pred.edges[b + 1]
                ghost_pred_b = ghost_pred[b] if ghost_pred is not None else None
                clust_label_adapted[lower:upper] = self._process(
                    clust_label[b], seg_label[b], seg_pred[b], ghost_pred_b
                )

            return TensorBatch(clust_label_adapted, seg_pred.counts)

        else:
            # Otherwise, call the main process function directly
            return self._process(clust_label, seg_label, seg_pred, ghost_pred)

    def _process(self, clust_label, seg_label, seg_pred, ghost_pred=None):
        """Adapts the cluster labels for one entry or a batch of entries.

        Parameters
        ----------
        clust_label : Union[np.ndarray, torch.Tensor]
            (N, N_l) Cluster label tensor
        seg_label : Union[np.ndarray, torch.Tensor]
            (M, 5) Segmentation label tensor
        seg_pred : Union[np.ndarray, torch.Tensor]
            (M/N_deghost) Segmentation predictions for each voxel
        ghost_pred : Union[np.ndarray, torch.Tensor], optional
            (M) Ghost predictions for each voxel

        Returns
        -------
        Union[np.ndarray, torch.Tensor]
            (N_deghost, N_l) Adapted cluster label tensor
        """
        # If there are no points in this event, nothing to do
        coords = seg_label[:, :VALUE_COL]
        num_cols = clust_label.shape[1]
        if not len(coords):
            return self._ones((0, num_cols))

        # If there are no points after deghosting, nothing to do
        if ghost_pred is not None:
            deghost_index = self._where(ghost_pred == 0)[0]
            if not len(deghost_index):
                return self._ones((0, num_cols))

        # If there are no label points in this event, return dummy labels
        if not len(clust_label):
            if ghost_pred is None:
                shape = (len(coords), num_cols)
                dummy_labels = -self._ones(shape)
                dummy_labels[:, :VALUE_COL] = coords

            else:
                shape = (len(deghost_index), num_cols)
                dummy_labels = -self._ones(shape)
                dummy_labels[:, :VALUE_COL] = coords[deghost_index]

            return dummy_labels

        # Build a tensor of predicted segmentation that includes ghost points
        seg_label = self._to_long(seg_label[:, SHAPE_COL])
        if ghost_pred is not None and (len(ghost_pred) != len(seg_pred)):
            seg_pred_long = self._to_long(GHOST_SHP * self._ones(len(coords)))
            seg_pred_long[deghost_index] = seg_pred
            seg_pred = seg_pred_long

        # Prepare new labels
        new_label = -self._ones((len(coords), num_cols))
        new_label[:, :VALUE_COL] = coords

        # Check if the segment labels and predictions are compatible. If they are
        # compatible, store the cluster labels as is. Track points do not mix
        # with other classes, but EM classes are allowed to.
        compat_mat = self._eye(GHOST_SHP + 1)
        compat_mat[
            (
                [SHOWR_SHP, SHOWR_SHP, MICHL_SHP, DELTA_SHP],
                [MICHL_SHP, DELTA_SHP, SHOWR_SHP, SHOWR_SHP],
            )
        ] = True

        true_deghost = seg_label < GHOST_SHP
        seg_mismatch = ~compat_mat[(seg_pred, seg_label)]
        new_label[true_deghost] = clust_label
        new_label[true_deghost & seg_mismatch, VALUE_COL:] = -self._ones(1)

        # For mismatched predictions, attempt to find a touching instance of the
        # same class to assign it sensible cluster labels.
        for s in self._unique(seg_pred):
            # Skip predicted ghosts (they keep their invalid labels)
            if s == GHOST_SHP:
                continue

            # Restrict to points in this class that have incompatible segment
            # labels. Track points do not mix, EM points are allowed to.
            bad_index = self._where((seg_pred == s) & (~true_deghost | seg_mismatch))[0]
            if len(bad_index) == 0:
                continue

            # Find points in clust_label that have compatible segment labels
            seg_clust_mask = compat_mat[s][self._to_long(clust_label[:, SHAPE_COL])]
            X_true = clust_label[seg_clust_mask]
            if len(X_true) == 0:
                continue

            # Loop over the set of unlabeled predicted points
            X_pred = coords[bad_index]
            tagged_voxels_count = 1
            while tagged_voxels_count > 0 and len(X_pred) > 0:
                # Compute Chebyshev distance between predicted and closest true
                distances = self._compute_distances(X_pred, X_true)
                distances, closest_ids = self._min(distances, 1)

                # Label unlabeled voxels that touch a compatible true voxel
                select_mask = distances < 1.1
                select_index = self._where(select_mask)[0]
                tagged_voxels_count = len(select_index)
                if tagged_voxels_count > 0:
                    # Use the label of the touching true voxel
                    additional_clust_label = self._cat(
                        [
                            X_pred[select_index],
                            X_true[closest_ids[select_index], VALUE_COL:],
                        ],
                        1,
                    )
                    new_label[bad_index[select_index]] = additional_clust_label

                    # Update the mask to not include the new assigned points
                    leftover_index = self._where(~select_mask)[0]
                    bad_index = bad_index[leftover_index]

                    # The new true available points are the ones we just added.
                    # The new pred points are those not yet labeled
                    X_true = additional_clust_label
                    X_pred = X_pred[leftover_index]

        # Remove predicted ghost points from the labels, set the shape
        # column of the label to the segmentation predictions.
        if ghost_pred is not None:
            new_label = new_label[deghost_index]
            new_label[:, SHAPE_COL] = seg_pred[deghost_index]
        else:
            new_label[:, SHAPE_COL] = seg_pred

        # Build a list of cluster indexes to break
        new_label_np = new_label
        if torch.is_tensor(new_label):
            new_label_np = new_label.detach().cpu().numpy()

        clusts = []
        labels = new_label_np[:, CLUST_COL]
        shapes = new_label_np[:, SHAPE_COL]
        for break_class in self.break_classes:
            index_s = np.where(shapes == break_class)[0]
            labels_s = labels[index_s]
            for c in np.unique(labels_s):
                # If the cluster ID is invalid, skip
                if c < 0:
                    continue

                # Append cluster
                clusts.append(index_s[labels_s == c])

        # Now if an instance was broken up, assign it different cluster IDs
        new_label[:, CLUST_COL] = self._offset + break_clusters(
            new_label, clusts, self.break_eps, self.break_metric_id, self.break_p
        )
        self._offset = new_label[:, CLUST_COL].max() + 1

        return new_label

    def _where(self, x):
        if self.torch:
            return torch.where(x)
        else:
            return np.where(x)

    def _cat(self, x, axis):
        if self.torch:
            return torch.cat(x, axis)
        else:
            return np.concatenate(x, axis)

    def _ones(self, x):
        if self.torch:
            return torch.ones(x, dtype=self.dtype, device=self.device)
        else:
            return np.ones(x)

    def _eye(self, x):
        if self.torch:
            return torch.eye(x, dtype=torch.bool, device=self.device)
        else:
            return np.eye(x, dtype=bool)

    def _min(self, x, axis):
        if self.torch:
            return torch.min(x, axis)
        else:
            return np.min(x, axis), np.argmin(x, axis)

    def _unique(self, x):
        if self.torch:
            return torch.unique(x).long()
        else:
            return np.unique(x).astype(np.int64)

    def _to_long(self, x):
        if self.torch:
            return x.long()
        else:
            return x.astype(int64)

    def _compute_distances(self, x, y):
        if self.torch:
            return cdist_fast(x[:, COORD_COLS], y[:, COORD_COLS], metric="chebyshev")
        else:
            return cdist(
                x[:, COORD_COLS], y[:, COORD_COLS], metric_id=METRICS["chebyshev"]
            )



"""Module with methods to augment the input data."""

import numpy as np

from spine.data import Meta

__all__ = ["Augmenter"]


class Augmenter:
    """Generic class to handle data augmentation."""

    def __init__(self, translate=None):
        """Initialize the augmenter.

        Parameters
        ----------
        translate : dict, optional
            Translation configuration (move input image around)
        """
        # Make sure at least one augmentation scheme is requested
        assert (
            translate is not None
        ), "Must provide `translate` block minimally to do any augmentation."

        # Parse the translation configuration
        self.translater = None
        if translate is not None:
            self.translater = Translater(**translate)

    def __call__(self, data):
        """Augment the data products in one event.

        Parameters
        ----------
        data : dict
           Data product dictionary
        """
        # Get the list of keys to augment and the shared metadata
        augment_keys = []
        meta = None
        for key, value in data.items():
            if (
                isinstance(value, tuple)
                and len(value) == 3
                and isinstance(value[2], Meta)
            ):
                augment_keys.append(key)
                if meta is None:
                    meta = value[2]
                else:
                    assert (
                        meta == value[2]
                    ), "Metadata should be shared by all data products."
            elif isinstance(value, Meta):
                augment_keys.append(key)
                meta = value

        # If there are no sparse tensors in the input data, nothing to do
        if meta is None:
            return data

        # Translate
        if self.translater is not None:
            data = self.translater(data, meta, augment_keys)

        return data


class Translater:
    """Generic class to handle moving images around."""

    def __init__(self, lower, upper):
        """Initialize the translater..

        This defines a way to move the image around within a volume greater
        than that define by the image metadata. The box must be larger than
        the image itself.

        Parameters
        ----------
        lower : np.ndarray
            Lower bounds of the box in which to move the image around
        upper : np.ndarray
            Upper bounds of the box in which to move the image around
        """
        # Sanity check
        assert (
            len(lower) == len(upper) == 3
        ), "Must provide boundaries for each dimension."

        # Define a new image metadata corresponding to the full range
        self.meta = Meta(lower=np.asarray(lower), upper=np.asarray(upper))

    def __call__(self, data, meta, keys):
        """Move an image around within the the pre-defined volume.

        Parameters
        ----------
        data : dict
            Dictionary of data products to offset
        meta : Meta
            Shared image metadata
        keys : List[str]
            List of keys with coordinates to offset

        Returns
        -------
        np.ndarray
            (N, 3) Translated points
        """
        # Set the target volume pixel pitch to match that of the original image
        if np.all(self.meta.size < 0.0):
            self.meta.size = meta.size
            self.meta.count = (self.meta.upper - self.meta.lower) // meta.size
            self.meta.count = self.meta.count.astype(int)

        # Generate an offset
        offset = self.generate_offset(meta)

        # Offset all coordinates
        for key in keys:
            # If the key is the metadata, modify and continue
            if isinstance(data[key], Meta):
                data[key] = self.meta
                continue

            # Fetch attributes to modify
            voxels, features, _ = data[key]

            # Translate
            width = voxels.shape[1]
            voxels = (voxels.reshape(-1, 3) + offset).reshape(-1, width)

            # Update
            data[key] = (voxels, features, self.meta)

        return data

    def generate_offset(self, meta):
        """Generate an offset to apply to all the voxel index sets.

        This offset is such that the the voxels will be randomly shifted
        within the target bounding box.

        Parameters
        ----------
        meta : Meta
            Metadata of the original image

        Returns
        -------
        np.ndarray
            Value by which to shift the pixels by
        """
        # Check that the original metadata is compatible with the target volume
        assert np.all(
            meta.count <= self.meta.count
        ), "The input image is larger than the target translation volume."

        # Generate an offset with respect to the voxel indices
        offset = np.random.randint((self.meta.count - meta.count) + 1)

        return offset



"""Module with a data class objects which represent output fragments."""

from dataclasses import dataclass

import numpy as np

from spine.data.particle import Particle
from spine.utils.docstring import inherit_docstring
from spine.utils.globals import SHAPE_LABELS, TRACK_SHP

from .base import RecoBase, TruthBase

__all__ = ["RecoFragment", "TruthFragment"]


@dataclass(eq=False)
class FragmentBase:
    """Base fragment-specific information.

    Attributes
    ----------
    particle_id : int
        Index of the particle this fragment belongs to
    interaction_id : int
        Index of the interaction this fragment belongs to
    shape : int
        Semantic type (shower (0), track (1), Michel (2), delta (3),
        low energy scatter (4)) of this particle
    is_primary : bool
        Whether this fragment was the first in the particle group
    length : float
        Length of the particle (only assigned to track objects)
    start_point : np.ndarray
        (3) Fragment start point
    end_point : np.ndarray
        (3) Fragment end point (only assigned to track objects)
    start_dir : np.ndarray
        (3) Fragment direction w.r.t. the start point
    end_dir : np.ndarray
        (3) Fragment direction w.r.t. the end point (only assigned
        to track objects)
    """

    particle_id: int = -1
    interaction_id: int = -1
    shape: int = -1
    is_primary: bool = False
    length: float = -1.0
    start_point: np.ndarray = None
    end_point: np.ndarray = None
    start_dir: np.ndarray = None
    end_dir: np.ndarray = None

    # Fixed-length attributes
    _fixed_length_attrs = (
        ("start_point", 3),
        ("end_point", 3),
        ("start_dir", 3),
        ("end_dir", 3),
    )

    # Attributes specifying coordinates
    _pos_attrs = ("start_point", "end_point")

    # Attributes specifying vector components
    _vec_attrs = ("start_dir", "end_dir")

    # Boolean attributes
    _bool_attrs = ("is_primary",)

    # Enumerated attributes
    _enum_attrs = (("shape", tuple((v, k) for k, v in SHAPE_LABELS.items())),)

    def __str__(self):
        """Human-readable string representation of the fragment object.

        Results
        -------
        str
            Basic information about the fragment properties
        """
        shape_label = SHAPE_LABELS[self.shape]
        match = self.match_ids[0] if len(self.match_ids) > 0 else -1
        return (
            f"Fragment(ID: {self.id:<3} | Shape: {shape_label:<11} "
            f"| Primary: {self.is_primary:<2} "
            f"| Size: {self.size:<5} | Match: {match:<3})"
        )


@dataclass(eq=False)
@inherit_docstring(RecoBase, FragmentBase)
class RecoFragment(FragmentBase, RecoBase):
    """Reconstructed fragment information.

    Attributes
    ----------
    primary_scores : np.ndarray
        (2) Array of softmax scores associated with secondary and primary
    """

    primary_scores: np.ndarray = None

    # Fixed-length attributes
    _fixed_length_attrs = (("primary_scores", 2), *FragmentBase._fixed_length_attrs)

    # Boolean attributes
    _bool_attrs = (*RecoBase._bool_attrs, *FragmentBase._bool_attrs)

    def __str__(self):
        """Human-readable string representation of the fragment object.

        Results
        -------
        str
            Basic information about the fragment properties
        """
        return "Reco" + super().__str__()


@dataclass(eq=False)
@inherit_docstring(TruthBase, FragmentBase)
class TruthFragment(Particle, FragmentBase, TruthBase):
    """Truth fragment information.

    This inherits all of the attributes of :class:`Particle`, which contains
    the G4 truth information for the fragment.

    Attributes
    ----------
    orig_interaction_id : int
        Unaltered index of the interaction in the original MC particle list
    orig_parent_id : int
        Unaltered index of the particle parent in the original MC particle list
    orig_group_id : int
        Unaltered index of the particle group in the original MC particle list
    orig_children_id : np.ndarray
        Unaltered list of the particle children in the original MC particle list
    children_counts : np.ndarray
        (P) Number of truth child fragment of each shape
    reco_length : float
        Reconstructed length of the fragment (only assigned to track objects)
    reco_start_dir : np.ndarray
        (3) Particle direction estimate w.r.t. the start point
    reco_end_dir : np.ndarray
        (3) Particle direction estimate w.r.t. the end point (only assigned
        to track objects)
    """

    orig_interaction_id: int = -1
    orig_parent_id: int = -1
    orig_group_id: int = -1
    orig_children_id: np.ndarray = None
    children_counts: np.ndarray = None
    reco_length: float = -1.0
    reco_start_dir: np.ndarray = None
    reco_end_dir: np.ndarray = None

    # Fixed-length attributes
    _fixed_length_attrs = (
        ("reco_start_dir", 3),
        ("reco_end_dir", 3),
        *FragmentBase._fixed_length_attrs,
        *Particle._fixed_length_attrs,
    )

    # Variable-length attributes
    _var_length_attrs = (
        ("orig_children_id", np.int64),
        ("children_counts", np.int32),
        *TruthBase._var_length_attrs,
        *Particle._var_length_attrs,
    )

    # Attributes specifying coordinates
    _pos_attrs = (*FragmentBase._pos_attrs, *Particle._pos_attrs)

    # Attributes specifying vector components
    _vec_attrs = (
        "reco_start_dir",
        "reco_end_dir",
        *FragmentBase._vec_attrs,
        *Particle._vec_attrs,
    )

    # Boolean attributes
    _bool_attrs = (*TruthBase._bool_attrs, *FragmentBase._bool_attrs)

    def __str__(self):
        """Human-readable string representation of the fragment object.

        Results
        -------
        str
            Basic information about the fragment properties
        """
        return "Truth" + super().__str__()

    @property
    def start_dir(self):
        """Converts the initial momentum to a direction vector.

        Returns
        -------
        np.ndarray
            (3) Start direction vector
        """
        if self.momentum is not None:
            norm = np.linalg.norm(self.momentum)
            if norm > 0.0 and norm != np.inf:
                return self.momentum / norm

        return np.full(3, -np.inf, dtype=np.float32)

    @start_dir.setter
    def start_dir(self, start_dir):
        pass

    @property
    def end_dir(self):
        """Converts the final momentum to a direction vector.

        Note that if a particle stops, this is unreliable as an estimate of the
        direction of the particle before it stops.

        Returns
        -------
        np.ndarray
            (3) End direction vector
        """
        if self.end_momentum is not None:
            norm = np.linalg.norm(self.end_momentum)
            if self.shape == TRACK_SHP and norm > 0.0 and norm != np.inf:
                return self.end_momentum / norm

        return np.full(3, -np.inf, dtype=np.float32)

    @end_dir.setter
    def end_dir(self, end_dir):
        pass



"""Tools to draw voxelized data organized in clusts."""

import time

import numpy as np

from .cone import cone_trace
from .ellipsoid import ellipsoid_trace
from .hull import hull_trace
from .point import scatter_points

__all__ = ["scatter_clusters"]


def scatter_clusters(
    points,
    clusts,
    color=None,
    hovertext=None,
    single_trace=False,
    name=None,
    mode="scatter",
    cmin=None,
    cmax=None,
    shared_legend=True,
    **kwargs,
):
    """Arranges points in clusters and scatters them and their cluster labels.

    Produces :class:`plotly.graph_objs.Scatter3d` trace object to be drawn
    in plotly. The object is nested to be fed directly to a
    :class:`plotly.graph_objs.Figure` or :func:`plotly.offline.iplot`.
    All of the regular plotly parameters are available.

    Parameters
    ----------
    points : np.ndarray
        (N, 3) array of N points of (..., x, y, z,...) coordinate information
    clusts : List[np.ndarray]
        (C) List of cluster indexes
    color : Union[str, np.ndarray], optional
        Color of markers or (N/C) list of color of markers or clusters
    hovertext : Union[List[str], List[int]], optional
        (N/C) List of labels associated with each marker or cluster
    single_trace : bool, default False
        If `True`, combine all clusters into a single plotly trace
    name : Union[str, List[str]], optional
        Name of the clusters or of each cluster
    mode : str, default 'scatter'
        Drawing mode; one of 'circle', 'scatter', 'ellipsoid', 'cone' or 'hull'
    cmin : float, optional
        Minimum value along the color scale
    cmax : float, optional
        Maximum value along the color scale
    shared_legend : bool, default True
        If `True` put all cluster traces under a single shared legend
    **kwargs : dict, optional
        List of additional arguments to pass to plotly.graph_objs.Scatter3D

    Returns
    -------
    List[go.Scatter3d]
        (1/C) List with one combined trace or one trace per cluster
    """
    # Build the point coordinate sets
    coords = [points[c] for c in clusts]

    # Get a single cluster index value per points
    counts = [len(c) for c in clusts]
    clust_ids = np.arange(len(clusts))

    # Build the color vectors
    has_labels = False
    if color is not None:
        has_labels = True
        if np.isscalar(color):
            color = [color] * len(clusts)
        elif len(color) == len(points) and len(points) != len(clusts):
            color = [color[c] for c in clusts]
        elif len(color) == len(clusts):
            if mode == "scatter" and len(color) > 0 and np.isscalar(color[0]):
                color = [[color[i]] * len(c) for i, c in enumerate(clusts)]
        else:
            raise ValueError(
                "The `color` attribute should be provided as a scalar, "
                "one value per point or one value per cluster."
            )

    else:
        if mode != "scatter":
            color = clust_ids
        else:
            color = [[clust_ids[i]] * len(c) for i, c in enumerate(clusts)]

    # Build the hovertext vectors
    if hovertext is not None:
        if np.isscalar(hovertext):
            hovertext = [hovertext] * len(clusts)
        elif len(hovertext) == len(points) and len(points) != len(clusts):
            hovertext = [hovertext[c] for c in clusts]
        elif len(hovertext) == len(hovertext):
            if mode == "scatter" and len(hovertext) > 0 and np.isscalar(hovertext[0]):
                hovertext = [[hovertext[i]] * len(c) for i, c in enumerate(clusts)]
        elif len(hovertext) != len(clusts):
            raise ValueError(
                "The `hovertext` attribute should be provided as a scalar, "
                "one value per point or one value per cluster."
            )

    else:
        hovertext = [f"Cluster ID: {i:.0f}" for i in clust_ids]
        if has_labels and len(color):
            if np.isscalar(color[0]):
                for i, hc in enumerate(hovertext):
                    fmt = ".0f" if float(color[i]).is_integer() else ".2f"
                    hovertext[i] = hc + f"<br>Label: {color[i]:{fmt}}"
            else:
                for i, hc in enumerate(hovertext):
                    hovertext[i] = [hc + f"<br>Value: {v:0.3f}" for v in color[i]]
        elif mode == "scatter":
            hovertext = [[hovertext[i]] * len(c) for i, c in enumerate(clusts)]

    # If requested, combine all clusters into a single trace
    if single_trace:
        # Check that we are operating in the expected mode
        assert mode in [
            "circle",
            "scatter",
        ], "Can only combine in one trace in 'circle' or 'scatter' mode."
        assert (
            shared_legend
        ), "Cannot split legend when merging all clusters in one trace."

        # Aggregate the coordinates, color and hovertext
        if mode == "circle":
            # Define the nodes as circles centered in the centroid of each
            # cluster and of radius proportional to the sqrt of the cluster size
            centroids = np.empty((len(coords), 3), dtype=np.float32)
            for i, coord in enumerate(coords):
                centroids[i] = np.mean(coord, axis=0)
            sizes = np.sqrt(counts)

            return scatter_points(
                centroids,
                name=name,
                color=color,
                markersize=sizes,
                hovertext=hovertext,
                cmin=cmin,
                cmax=cmax,
                **kwargs,
            )

        else:
            if len(coords):
                coords = np.vstack(coords)
            else:
                coords = np.empty((0, 3), dtype=np.float32)

            if color is not None and len(color):
                color = np.concatenate(color)
            if hovertext is not None and len(hovertext):
                hovertext = np.concatenate(hovertext)

            return scatter_points(
                coords,
                color=color,
                hovertext=hovertext,
                name=name,
                cmin=cmin,
                cmax=cmax,
                **kwargs,
            )

    # If cmin/cmax are not provided, must build them so that all clusters
    # share the same colorscale range (not guaranteed otherwise)
    if color is not None and not np.isscalar(color) and len(color) > 0:
        if cmin is None:
            if np.isscalar(color[0]):
                cmin = np.min(color)
            else:
                cmin = np.min(np.concatenate(color))

        if cmax is None:
            if np.isscalar(color[0]):
                cmax = np.max(color)
            else:
                cmax = np.max(np.concatenate(color))

    # Loop over the list of clusters
    traces = []
    group_name = "group_" + str(time.time())
    for i, coord in enumerate(coords):
        # If the legend is shared, only draw the legend of the first trace
        legendgroup, showlegend, name_i = None, True, name
        if shared_legend:
            legendgroup = group_name
            showlegend = i == 0
        elif name is not None:
            if np.isscalar(name):
                name_i = f"{name} {i}"
            else:
                assert len(name) == len(clusts), (
                    "When providing the name as a list, there should be "
                    "one name per cluster."
                )
                name_i = name[i]

        # Dispatch
        if mode == "circle":
            centroid = np.mean(coord, axis=0)[None, :]
            size = np.sqrt(counts[i])
            traces += scatter_points(
                centroid,
                name=name_i,
                color=color[i],
                hovertext=hovertext[i],
                cmin=cmin,
                cmax=cmax,
                markersize=size,
                legendgroup=legendgroup,
                showlegend=showlegend,
                **kwargs,
            )

        elif mode == "scatter":
            traces += scatter_points(
                coord,
                name=name_i,
                color=color[i],
                hovertext=hovertext[i],
                cmin=cmin,
                cmax=cmax,
                legendgroup=legendgroup,
                showlegend=showlegend,
                **kwargs,
            )

        elif mode == "ellipsoid":
            traces.append(
                ellipsoid_trace(
                    coord,
                    name=name_i,
                    color=color[i],
                    hovertext=hovertext[i],
                    cmin=cmin,
                    cmax=cmax,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    **kwargs,
                )
            )

        elif mode == "cone":
            traces.append(
                cone_trace(
                    coord,
                    name=name_i,
                    color=color[i],
                    hovertext=hovertext[i],
                    cmin=cmin,
                    cmax=cmax,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    **kwargs,
                )
            )

        elif mode == "hull":
            traces.append(
                hull_trace(
                    coord,
                    name=name_i,
                    color=color[i],
                    hovertext=hovertext[i],
                    cmin=cmin,
                    cmax=cmax,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    **kwargs,
                )
            )

        else:
            raise ValueError(
                f"Cluster drawing mode not recognized: {mode}. Must be one "
                "of 'circle', 'sctatter', 'ellipsoid', 'cone' or 'hull'."
            )

    return traces



"""Analysis module template.

Use this template as a basis to build your own analysis script. An analysis
script takes the output of the reconstruction and the post-processors and
performs basic selection cuts and store the output to a CSV file.
"""

# Add the imports specific to this module here
# import ...

# Must import the analysis script base class
from spine.ana.base import AnaBase

# Must list the analysis script(s) here to be found by the factory.
# You must also add it to the list of imported modules in the
# `spine.ana.factories`!
__all__ = ["TemplateAna"]


class TemplateAna(AnaBase):
    """Description of what the analysis script is supposed to be doing."""

    # Name of the analysis script (as specified in the configuration)
    name = "template"

    def __init__(
        self, arg0, arg1, obj_type, run_mode, append_file, overwrite_file, output_prefix
    ):
        """Initialize the analysis script.

        Parameters
        ----------
        arg0 : type
            Description of arg0
        arg1 : type
            Description of arg1
        obj_type : Union[str, List[str]]
            Name or list of names of the object types to process
        run_mode : str, optional
            If specified, tells whether the analysis script must run on
            reconstructed ('reco'), true ('true') or both objects
            ('both' or 'all')
        append_file : bool, default False
            If True, appends existing CSV files instead of creating new ones
        overwrite_file : bool, default False
            If True and the output CSV file exists, overwrite it
        output_prefix : str, default None
            Name to prefix every output CSV file with
        """
        # Initialize the parent class
        super().__init__(obj_type, run_mode, append_file, overwrite_file, output_prefix)

        # Store parameter
        self.arg0 = arg0
        self.arg1 = arg1

        # Initialize the CSV writer(s) you want
        self.initialize_writer("template")

        # Add additional required data products
        self.update_keys({"prod": True})  # Means we must have 'prod' in the dictionary

    def process(self, data):
        """Pass data products corresponding to one entry through the analysis.

        Parameters
        ----------
        data : dict
            Dictionary of data products
        """
        # Fetch the keys you want
        data = data["prod"]

        # Loop over all requested object types
        for key in self.obj_keys:
            # Loop over all objects of that type
            for obj in data[key]:
                # Do something with the object
                disp = obj.end_point - obj.start_point

                # Make a dictionary of integer out of it
                out = {"disp_x": disp[0], "disp_y": disp[1], "disp_z": disp[2]}

                # Write the row to file
                self.append("template", **out)



"""Functions that instantiate PyTorch-independent IO tools from configuration blocks."""

from warnings import warn

from spine.utils.factory import instantiate, module_dict

from . import read, write

READER_DICT = module_dict(read)
WRITER_DICT = module_dict(write)

__all__ = ["reader_factory", "writer_factory"]


def reader_factory(reader_cfg):
    """Instantiates reader based on type specified in configuration under
    `io.reader.name`. The name must match the name of a class under
    `spine.io.readers`.

    Parameters
    ----------
    reader_cfg : dict
        Writer configuration dictionary

    Returns
    -------
    object
        Writer object

    Note
    ----
    Currently the choice is limited to `HDF5Writer` only.
    """
    # Initialize reader
    return instantiate(READER_DICT, reader_cfg)


def writer_factory(writer_cfg, prefix=None, split=False):
    """Instantiates writer based on type specified in configuration under
    `io.writer.name`. The name must match the name of a class under
    `spine.io.writers`.

    Parameters
    ----------
    writer_cfg : dict
        Writer configuration dictionary
    prefix : str, optional
        Input file prefix to use as an output name
    split : bool, default False
        Split the output into one file per input file

    Returns
    -------
    object
        Writer object

    Note
    ----
    Currently the choice is limited to `HDF5Writer` only.
    """
    # Initialize writer
    return instantiate(WRITER_DICT, writer_cfg, prefix=prefix, split=split)



"""Delaunay graph constructor for GNNs."""

import numba as nb
import numpy as np
from scipy.spatial import Delaunay

from spine.utils.globals import COORD_COLS

from .base import GraphBase

__all__ = ["DelaunayGraph"]


class DelaunayGraph(GraphBase):
    """Generates graphs based on the Delaunay triangulation of the input
    node locations.

    Triangulates the input, converts the triangles into a list of valid edges.

    See :class:`GraphBase` for attributes/methods shared
    across all graph constructors.
    """

    # Name of the graph constructor (as specified in the configuration)
    name = "delaunay"

    def generate(self, data, clusts, **kwargs):
        """Generates an incidence matrix that connects nodes
        that share an edge in their corresponding Euclidean Delaunay graph.

        Parameters
        ----------
        data : TensorBatch
            (N, 1 + D + N_f) Tensor of voxel/value pairs
        clusts : IndexBatch
            (C) Cluster indexes
        **kwargs : dict, optional
            Unused graph generation arguments

        Returns
        -------
        np.ndarray
            (2, E) Tensor of edges
        """
        return self._generate(
            data.tensor,
            nb.typed.List(clusts.index_list),
            clusts.batch_ids,
            self.directed,
        )

    @staticmethod
    @nb.njit(cache=True)
    def _generate(
        data: nb.float64[:, :],
        clusts: nb.types.List(nb.int64[:]),
        batch_ids: nb.int64[:],
        directed: bool = False,
    ) -> nb.int64[:, :]:
        # For each batch, find the list of edges, append it
        edge_list, offset = [], 0
        edge_counts = np.zeros(len(counts), dtype=counts.dtype)
        for b in np.unique(batch_ids):
            # Combine the cluster masks into one
            clust_ids = np.where(batch_ids == b)[0]
            limits = np.array([0] + [len(clusts[i]) for i in clust_ids])
            limits = np.cumsum(limits)
            mask = np.zeros(limits[-1], dtype=np.int64)
            labels = np.zeros(limits[-1], dtype=np.int64)
            for i in range(len(clust_ids)):
                l, u = limits[i], limits[i + 1]
                mask[l:u] = clusts[clust_ids[i]]
                labels[l:u] = i

            # Run Delaunay triangulation in object mode because it relies on an
            # external package. Only way to speed this up would be to implement
            # Delaunay triangulation in Numba (tall order)
            with nb.objmode(tri="int32[:,:]"):
                # Run Delaunay triangulation in joggled mode, this
                # guarantees simplical faces (no ambiguities)
                tri = Delaunay(data[mask][:, COORD_COLS], qhull_options="QJ").simplices

            # Create an adjanceny matrix from the simplex list
            adj_mat = np.zeros((len(clust_ids), len(clust_ids)), dtype=np.bool_)
            for s in tri:
                for i in s:
                    for j in s:
                        if labels[j] > labels[i]:
                            adj_mat[labels[i], labels[j]] = True

            # Convert the adjancency matrix to a list of edges, store
            edges = np.where(adj_mat)
            edges = np.vstack((clust_ids[edges[0]], clust_ids[edges[1]]))

            edge_list.append(offset + edge_index.T)
            edge_counts[b] = edge_index.shape[-1]
            offset += c

        # Merge the blocks together
        offset = 0
        result = np.empty((2, num_edges), dtype=np.int64)
        for edges in edge_list:
            num_block = edges.shape[-1]
            result[:, offset : offset + num_block] = edges
            offset += num_block

        return result
